# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""quota_request supports aggregation of AllocateQuotaRequests.

:func:`sign` generated a signature from AllocateQuotaRequests
:class:`~endpoints_management.gen.servicecontrol_v1_message.QuotaOperation` represents
information regarding an operation, and is a key constituent of
:class:`~endpoints_management.gen.servicecontrol_v1_message.AllocateQuotaRequest`.

The :class:`.Aggregator` implements the strategy for aggregating AllocateQuotaRequests
and caching their responses.

"""

from __future__ import absolute_import

import collections
import copy
import hashlib
import httplib
import logging
from datetime import datetime

from apitools.base.py import encoding

from . import (caches, label_descriptor, metric_value, operation, sc_messages,
               signing)
from .. import USER_AGENT, SERVICE_AGENT

_logger = logging.getLogger(__name__)

# alias for brevity
_QuotaErrors = sc_messages.QuotaError.CodeValueValuesEnum
_IS_OK = (httplib.OK, u'')
_IS_UNKNOWN = (
    httplib.INTERNAL_SERVER_ERROR,
    u'Request blocked due to unsupported block reason {detail}')
_QUOTA_ERROR_CONVERSION = {
    _QuotaErrors.RESOURCE_EXHAUSTED: (
        429,
        'Quota allocation failed',
    ),
    _QuotaErrors.BILLING_NOT_ACTIVE: (
        httplib.FORBIDDEN,
        u'Project {project_id} has billing disabled. Please enable it',
    ),
    _QuotaErrors.PROJECT_DELETED: (
        httplib.FORBIDDEN,
        u'Project {project_id} has been deleted',
    ),
    _QuotaErrors.API_KEY_INVALID: (
        httplib.BAD_REQUEST,
        u'API not valid. Please pass a valid API key',
    ),
    _QuotaErrors.API_KEY_EXPIRED: (
        httplib.BAD_REQUEST,
        u'API key expired. Please renew the API key',
    ),


    # Fail open for internal server errors
    _QuotaErrors.UNSPECIFIED: _IS_OK,
    _QuotaErrors.PROJECT_STATUS_UNAVAILABLE: _IS_OK,
    _QuotaErrors.SERVICE_STATUS_UNAVAILABLE: _IS_OK,
    _QuotaErrors.BILLING_STATUS_UNAVAILABLE: _IS_OK,
    _QuotaErrors.QUOTA_SYSTEM_UNAVAILABLE: _IS_OK,
}


def convert_response(allocate_quota_response, project_id):
    """Computes a http status code and message `AllocateQuotaResponse`

    The return value a tuple (code, message) where

    code: is the http status code
    message: is the message to return

    Args:
       allocate_quota_response (:class:`endpoints_management.gen.servicecontrol_v1_messages.AllocateQuotaResponse`):
         the response from calling an api

    Returns:
       tuple(code, message)
    """
    if not allocate_quota_response or not allocate_quota_response.allocateErrors:
        return _IS_OK

    # only allocate_quota the first error for now, as per ESP
    theError = allocate_quota_response.allocateErrors[0]
    error_tuple = _QUOTA_ERROR_CONVERSION.get(theError.code, _IS_UNKNOWN)
    if error_tuple[1].find(u'{') == -1:  # no replacements needed:
        return error_tuple

    updated_msg = error_tuple[1].format(project_id=project_id, detail=theError.description or u'')
    return error_tuple[0], updated_msg


def sign(allocate_quota_request):
    """Obtains a signature for an operation in a `AllocateQuotaRequest`

    Args:
       op (:class:`endpoints_management.gen.servicecontrol_v1_messages.Operation`): an
         operation used in a `AllocateQuotaRequest`

    Returns:
       string: a secure hash generated from the operation
    """
    if not isinstance(allocate_quota_request, sc_messages.AllocateQuotaRequest):
        raise ValueError(u'Invalid request')
    op = allocate_quota_request.allocateOperation
    if op is None or op.methodName is None or op.consumerId is None:
        logging.error(u'Bad %s: not initialized => not signed', allocate_quota_request)
        raise ValueError(u'allocate_quota request must be initialized with an operation')
    md5 = hashlib.md5()
    md5.update(op.methodName.encode('utf-8'))
    md5.update(b'\x00')
    md5.update(op.consumerId.encode('utf-8'))
    if op.labels:
        signing.add_dict_to_hash(md5, encoding.MessageToPyValue(op.labels))
    for value_set in op.quotaMetrics:
        md5.update(b'\x00')
        md5.update(value_set.metricName.encode('utf-8'))
        for mv in value_set.metricValues:
            metric_value.update_hash(md5, mv)

    md5.update(b'\x00')
    return md5.digest()


_KNOWN_LABELS = label_descriptor.KnownLabels


_INFO_FIELDS = (u'client_ip', u'quota_info', u'config_id') + operation.Info._fields


class Info(collections.namedtuple(u'Info', _INFO_FIELDS), operation.Info):
    """Holds the information necessary to fill in QuotaRequest.

    In addition the attributes in :class:`operation.Info`, this has:

    Attributes:
       client_ip: the client IP address
       quota_info: the quota info from the method

    """
    def __new__(cls, client_ip=u'', quota_info=None, config_id=None, **kw):
        """Invokes the base constructor with default values."""
        op_info = operation.Info(**kw)
        return super(Info, cls).__new__(cls, client_ip, quota_info, config_id, **op_info._asdict())

    def as_allocate_quota_request(self, timer=datetime.utcnow):
        """Makes a `ServicecontrolServicesAllocateQuotaRequest` from this instance

        Returns:
          a ``ServicecontrolServicesAllocateQuotaRequest``

        Raises:
          ValueError: if the fields in this instance are insufficient to
            to create a valid ``ServicecontrolServicesAllocateQuotaRequest``

        """
        if not self.service_name:
            raise ValueError(u'the service name must be set')
        if not self.operation_id:
            raise ValueError(u'the operation id must be set')
        if not self.operation_name:
            raise ValueError(u'the operation name must be set')
        op = super(Info, self).as_operation(timer=timer)
        labels = {}
        if self.client_ip:
            labels[_KNOWN_LABELS.SCC_CALLER_IP.label_name] = self.client_ip

        if self.referer:
            labels[_KNOWN_LABELS.SCC_REFERER.label_name] = self.referer

        qop = sc_messages.QuotaOperation(
            operationId=op.operationId,
            methodName=op.operationName,
            consumerId=op.consumerId,
            quotaMode=sc_messages.QuotaOperation.QuotaModeValueValuesEnum.BEST_EFFORT,
        )
        qop.labels = encoding.PyValueToMessage(
            sc_messages.QuotaOperation.LabelsValue, labels)

        quota_info = self.quota_info if self.quota_info else {}
        qop.quotaMetrics = [
            sc_messages.MetricValueSet(
                metricName=name, metricValues=[sc_messages.MetricValue(int64Value=cost)])
            for name, cost in quota_info.items()
        ]

        allocate_quota_request = sc_messages.AllocateQuotaRequest(allocateOperation=qop)
        if self.config_id:
            allocate_quota_request.serviceConfigId = self.config_id
        return sc_messages.ServicecontrolServicesAllocateQuotaRequest(
            serviceName=self.service_name,
            allocateQuotaRequest=allocate_quota_request)


class Aggregator(object):
    """Caches and aggregates ``AllocateQuotaRequests``.

    Concurrency: Thread safe.

    """

    def __init__(self, service_name, options, kinds=None,
                 timer=datetime.utcnow):
        """Constructor.

        Args:
          service_name (string): names the service that all requests aggregated
            by this instance will be sent
          options (:class:`~endpoints_management.caches.QuotaOptions`): configures the
            caching and flushing behavior of this instance
          kinds (dict[string,[endpoints_management.control.MetricKind]]): specifies the
            kind of metric for each each metric name.
          timer (function([[datetime]]): a function that returns the current
            as a time as a datetime instance
        """
        self._service_name = service_name
        self._options = options
        self._cache = caches.create(options, timer=timer, use_deque=False)
        # When using the result of `with self._out as out`, you must disable no-member
        # in pyflakes. Known issue with no fix ETA:
        # https://github.com/PyCQA/astroid/issues/347
        # https://github.com/PyCQA/pylint/issues/1437
        self._out = caches.LockedObject(collections.deque())
        self._kinds = {} if kinds is None else dict(kinds)
        self._timer = timer
        self._in_flush_all = False

    @property
    def service_name(self):
        """The service to which all aggregated requests should belong."""
        return self._service_name

    @property
    def flush_interval(self):
        """The interval between calls to flush.

        Returns:
           timedelta: the period between calls to flush if, or ``None`` if no
           cache is set

        """
        return None if self._cache is None else self._options.flush_interval

    def flush(self):
        """Flushes this instance's cache.

        The driver of this instance should call this method every
        `flush_interval`.

        Returns:
          list['ServicecontrolServicesAllocateQuotaRequest']: corresponding
          to AllocateQuotaRequests that were pending

        """
        if self._cache is None:
            return []
        with self._cache as c, self._out as out:
            c.expire()
            now = self._timer()
            for item in c.values():
                if (not self._in_flush_all) and (not self._should_expire(item)):
                    if (not item.is_in_flight) and item._op_aggregator is not None:
                        item.is_in_flight = True
                        item.last_refresh_timestamp = now
                        out.append(item.extract_request())  # pylint: disable=no-member
            flushed_items = list(out)
            out.clear()  # pylint: disable=no-member
            for req in flushed_items:
                assert isinstance(req, sc_messages.ServicecontrolServicesAllocateQuotaRequest)
            return flushed_items

    def clear(self):
        """Clears this instance's cache."""
        if self._cache is not None:
            with self._cache as c, self._out as out:
                self.in_flush_all = True
                c.clear()
                out.clear()  # pylint: disable=no-member
                self.in_flush_all = False

    def add_response(self, req, resp):
        """Adds the response from sending to `req` to this instance's cache.

        Args:
          req (`ServicecontrolServicesAllocateQuotaRequest`): the request
          resp (AllocateQuotaResponse): the response from sending the request
        """
        if self._cache is None:
            return
        signature = sign(req.allocateQuotaRequest)
        with self._cache as c:
            now = self._timer()
            item = c.get(signature)
            if item is None:
                c[signature] = CachedItem(
                    req, resp, self.service_name, now)
            else:
                # Update the cached item to reflect that it is updated
                item.last_check_time = now
                item.response = resp
                item.is_in_flight = False
                c[signature] = item

    def allocate_quota(self, req):
        if self._cache is None:
            return None  # no cache, send request now
        if not isinstance(req, sc_messages.ServicecontrolServicesAllocateQuotaRequest):
            raise ValueError(u'Invalid request')
        if req.serviceName != self.service_name:
            _logger.error(u'bad allocate_quota(): service_name %s does not match ours %s',
                          req.serviceName, self.service_name)
            raise ValueError(u'Service name mismatch')
        allocate_quota_request = req.allocateQuotaRequest
        if allocate_quota_request is None:
            _logger.error(u'bad allocate_quota(): no allocate_quota_request in %s', req)
            raise ValueError(u'Expected operation not set')
        op = allocate_quota_request.allocateOperation
        if op is None:
            _logger.error(u'bad allocate_quota(): no operation in %s', req)
            raise ValueError(u'Expected operation not set')

        signature = sign(allocate_quota_request)
        with self._cache as cache, self._out as out:
            now = self._timer()
            _logger.debug(u'checking the cache for %r\n%s', signature, cache)
            item = cache.get(signature)
            if item is None:
                # to avoid sending concurrent allocate_quota from
                # concurrent requests, insert a temporary positive
                # response in the cache. Quota requests from other API
                # requests will be aggregated to this temporary
                # element until the response for the actual request
                # arrives.
                temp_response = sc_messages.AllocateQuotaResponse(operationId=op.operationId)
                item = CachedItem(allocate_quota_request, temp_response, self.service_name, now)
                item.signature = signature
                item.is_in_flight = True
                cache[signature] = item
                out.append(req)  # pylint: disable=no-member
                return temp_response  # positive response
            if not item.is_in_flight and self._should_refresh(item):
                item.is_in_flight = True
                item.last_refreshed_timestamp = now

                refresh_request = item.extract_request()
                if not item.is_positive_response():
                    # if the cached response is negative, then use NORMAL QuotaMode instead of BEST_EFFORT
                    normal = sc_messages.QuotaOperation.QuotaModeValueValuesEnum.NORMAL
                    refresh_request.allocateQuotaRequest.allocateOperation.quotaMode = normal
                out.append(refresh_request)  # pylint: disable=no-member
            if item.is_positive_response():
                item.aggregate(allocate_quota_request)
            return item.response

    def _should_refresh(self, item):
        age = self._timer() - item.last_check_time
        return age >= self._options.flush_interval

    def _should_expire(self, item):
        age = self._timer() - item.last_check_time
        return age >= self._options.expiration


class CachedItem(object):
    """CachedItem holds items cached along with a ``AllocateQuotaRequest``.

    Thread compatible.

    Attributes:
       response (:class:`sc_messages.CachedResponse`): the cached response
       is_flushing (bool): indicates if it's been detected that item
         is stale, and needs to be flushed
       quota_scale (int): WIP, used to determine quota
       last_check_time (datetime.datetime): the last time this instance
         was checked

    """

    def __init__(self, req, resp, service_name, last_check_time):
        assert isinstance(req, sc_messages.AllocateQuotaRequest)
        self.request = req
        self.response = resp
        self.last_check_time = last_check_time
        self.is_in_flight = False
        self._service_name = service_name
        self._op_aggregator = None

    def aggregate(self, req):
        assert isinstance(req, sc_messages.AllocateQuotaRequest)
        if self._op_aggregator is None:
            # This is correct behavior; we have no need to aggregate the req
            # from the constructor since it has already been sent.
            self._op_aggregator = QuotaOperationAggregator(req.allocateOperation)
        else:
            self._op_aggregator.merge_operation(req.allocateOperation)

    def extract_request(self):
        if self._op_aggregator is None:
            allocate_quota_request = self.request
        else:
            op = self._op_aggregator.as_quota_operation()
            self._op_aggregator = None
            allocate_quota_request = sc_messages.AllocateQuotaRequest(allocateOperation=op)
        return sc_messages.ServicecontrolServicesAllocateQuotaRequest(
            serviceName=self._service_name,
            allocateQuotaRequest=allocate_quota_request)

    def is_positive_response(self):
        return len(self.response.allocateErrors) == 0


class QuotaOperationAggregator(object):
    def __init__(self, op):
        # The protorpc version used here lacks MergeFrom
        self.op = copy.deepcopy(op)
        self.op.quotaMetrics = []
        self.metric_value_sets = {}
        self.merge_operation(op)

    def merge_operation(self, op):
        assert isinstance(op, sc_messages.QuotaOperation)
        for mv_set in op.quotaMetrics:
            metric_name = mv_set.metricName
            if metric_name not in self.metric_value_sets:
                self.metric_value_sets[metric_name] = mv_set.metricValues[0]
            else:
                self.metric_value_sets[metric_name] = metric_value.merge(
                    metric_value.MetricKind.DELTA,
                    self.metric_value_sets[metric_name],
                    mv_set.metricValues[0]
                )

    def as_quota_operation(self):
        op = copy.deepcopy(self.op)
        for m_name, m_value in self.metric_value_sets.items():
            op.quotaMetrics.append(sc_messages.MetricValueSet(
                metricName=m_name, metricValues=[m_value]))
        return op



# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""caches provide functions and classes used to support caching.

caching is provide by extensions of the cache classes provided by the
cachetools open-source library.

:func:`create` creates a cache instance specifed by either
:class:`endpoints_management.control.CheckAggregationOptions` or a
:class:`endpoints_management.control.ReportAggregationOptions`

"""

from __future__ import absolute_import

# pylint: disable=too-many-ancestors
#
# It affects the DequeOutTTLCache and DequeOutLRUCache which extend
# cachetools.TTLCache and cachetools.LRUCache respectively.  Within cachetools,
# those classes each extend Cache, which itself extends DefaultMapping. It does
# makes sense to have this chain of ancestors, so it's right the disable the
# warning here.

import collections
import logging
import threading
from datetime import datetime, timedelta

import cachetools

_logger = logging.getLogger(__name__)


class CheckOptions(
        collections.namedtuple(
            u'CheckOptions',
            [u'num_entries',
             u'flush_interval',
             u'expiration'])):
    """Holds values used to control report check behavior.

    Attributes:

        num_entries: the maximum number of cache entries that can be kept in
          the aggregation cache
        flush_interval (:class:`datetime.timedelta`): the maximum delta before
          aggregated report requests are flushed to the server.  The cache
          entry is deleted after the flush.
        expiration (:class:`datetime.timedelta`): elapsed time before a cached
          check response should be deleted.  This value should be larger than
          ``flush_interval``, otherwise it will be ignored, and instead a value
          equivalent to flush_interval + 1ms will be used.
    """
    # pylint: disable=too-few-public-methods
    DEFAULT_NUM_ENTRIES = 200
    DEFAULT_FLUSH_INTERVAL = timedelta(milliseconds=500)
    DEFAULT_EXPIRATION = timedelta(seconds=1)

    def __new__(cls,
                num_entries=DEFAULT_NUM_ENTRIES,
                flush_interval=DEFAULT_FLUSH_INTERVAL,
                expiration=DEFAULT_EXPIRATION):
        """Invokes the base constructor with default values."""
        assert isinstance(num_entries, int), u'should be an int'
        assert isinstance(flush_interval, timedelta), u'should be a timedelta'
        assert isinstance(expiration, timedelta), u'should be a timedelta'
        if expiration <= flush_interval:
            expiration = flush_interval + timedelta(milliseconds=1)
        return super(cls, CheckOptions).__new__(
            cls,
            num_entries,
            flush_interval,
            expiration)


class QuotaOptions(
        collections.namedtuple(
            u'QuotaOptions',
            [u'num_entries',
             u'flush_interval',
             u'expiration'])):
    """Holds values used to control report quota behavior.

    Attributes:

        num_entries: the maximum number of cache entries that can be kept in
          the aggregation cache
        flush_interval (:class:`datetime.timedelta`): the maximum delta before
          aggregated report requests are flushed to the server.  The cache
          entry is deleted after the flush.
        expiration (:class:`datetime.timedelta`): elapsed time before a cached
          quota response should be deleted.  This value should be larger than
          ``flush_interval``, otherwise it will be ignored, and instead a value
          equivalent to flush_interval + 1ms will be used.
    """
    # pylint: disable=too-few-public-methods
    DEFAULT_NUM_ENTRIES = 1000
    DEFAULT_FLUSH_INTERVAL = timedelta(seconds=1)
    DEFAULT_EXPIRATION = timedelta(minutes=1)

    def __new__(cls,
                num_entries=DEFAULT_NUM_ENTRIES,
                flush_interval=DEFAULT_FLUSH_INTERVAL,
                expiration=DEFAULT_EXPIRATION):
        """Invokes the base constructor with default values."""
        assert isinstance(num_entries, int), u'should be an int'
        assert isinstance(flush_interval, timedelta), u'should be a timedelta'
        assert isinstance(expiration, timedelta), u'should be a timedelta'
        if expiration <= flush_interval:
            expiration = flush_interval + timedelta(milliseconds=1)
        return super(cls, QuotaOptions).__new__(
            cls,
            num_entries,
            flush_interval,
            expiration)


class ReportOptions(
        collections.namedtuple(
            u'ReportOptions',
            [u'num_entries',
             u'flush_interval'])):
    """Holds values used to control report aggregation behavior.

    Attributes:

        num_entries: the maximum number of cache entries that can be kept in
          the aggregation cache

        flush_interval (:class:`datetime.timedelta`): the maximum delta before
          aggregated report requests are flushed to the server.  The cache
          entry is deleted after the flush
    """
    # pylint: disable=too-few-public-methods
    DEFAULT_NUM_ENTRIES = 200
    DEFAULT_FLUSH_INTERVAL = timedelta(seconds=1)

    def __new__(cls,
                num_entries=DEFAULT_NUM_ENTRIES,
                flush_interval=DEFAULT_FLUSH_INTERVAL):
        """Invokes the base constructor with default values."""
        assert isinstance(num_entries, int), u'should be an int'
        assert isinstance(flush_interval, timedelta), u'should be a timedelta'

        return super(cls, ReportOptions).__new__(
            cls,
            num_entries,
            flush_interval)


ZERO_INTERVAL = timedelta()


def create(options, timer=None, use_deque=True):
    """Create a cache specified by ``options``

    ``options`` is an instance of either
    :class:`endpoints_management.control.caches.CheckOptions` or
    :class:`endpoints_management.control.caches.ReportOptions`

    The returned cache is wrapped in a :class:`LockedObject`, requiring it to
    be accessed in a with statement that gives synchronized access

    Example:
      >>> options = CheckOptions()
      >>> synced_cache = make_cache(options)
      >>> with synced_cache as cache:  #  acquire the lock
      ...    cache['a_key'] = 'a_value'

    Args:
      options (object): an instance of either of the options classes

    Returns:
      :class:`cachetools.Cache`: the cache implementation specified by options
        or None: if options is ``None`` or if options.num_entries < 0

    Raises:
       ValueError: if options is not a support type

    """
    if options is None:  # no options, don't create cache
        return None

    if not isinstance(options, (CheckOptions, QuotaOptions, ReportOptions)):
        _logger.error(u'make_cache(): bad options %s', options)
        raise ValueError(u'Invalid options')

    if (options.num_entries <= 0):
        _logger.debug(u"did not create cache, options was %s", options)
        return None

    _logger.debug(u"creating a cache from %s", options)
    if (options.flush_interval > ZERO_INTERVAL):
        # options always has a flush_interval, but may have an expiration
        # field. If the expiration is present, use that instead of the
        # flush_interval for the ttl
        ttl = getattr(options, u'expiration', options.flush_interval)
        cache_cls = DequeOutTTLCache if use_deque else cachetools.TTLCache
        return LockedObject(
            cache_cls(
                options.num_entries,
                ttl=ttl.total_seconds(),
                timer=to_cache_timer(timer)
            ))

    cache_cls = DequeOutLRUCache if use_deque else cachetools.LRUCache
    return LockedObject(cache_cls(options.num_entries))


class DequeOutTTLCache(cachetools.TTLCache):
    """Extends ``TTLCache`` so that expired items are placed in a ``deque``."""

    def __init__(self, maxsize, ttl, out_deque=None, **kw):
        """Constructor.

        Args:
          maxsize (int): the maximum number of entries in the queue
          ttl (int): the ttl for entries added to the cache
          out_deque :class:`collections.deque`: a `deque` in which to add items
            that expire from the cache
          **kw: the other keyword args supported by the constructor to
            :class:`cachetools.TTLCache`

        Raises:
          ValueError: if out_deque is not a collections.deque

        """
        super(DequeOutTTLCache, self).__init__(maxsize, ttl, **kw)
        if out_deque is None:
            out_deque = collections.deque()
        elif not isinstance(out_deque, collections.deque):
            raise ValueError(u'out_deque should be a collections.deque')
        self._out_deque = out_deque
        self._tracking = {}

    def __setitem__(self, key, value, **kw):
        super(DequeOutTTLCache, self).__setitem__(key, value, **kw)
        self._tracking[key] = value

    @property
    def out_deque(self):
        """The :class:`collections.deque` to which expired items are added."""
        self.expire()
        expired = {k: v for (k, v) in self._tracking.items() if self.get(k) is None}
        for k, v in expired.items():
            del self._tracking[k]
            self._out_deque.append(v)
        return self._out_deque


class DequeOutLRUCache(cachetools.LRUCache):
    """Extends ``LRUCache`` so that expired items are placed in a ``deque``."""

    def __init__(self, maxsize, out_deque=None, **kw):
        """Constructor.

        Args:
          maxsize (int): the maximum number of entries in the queue
          out_deque :class:`collections.deque`: a `deque` in which to add items
            that expire from the cache
          **kw: the other keyword args supported by constructor to
            :class:`cachetools.LRUCache`

        Raises:
          ValueError: if out_deque is not a collections.deque

        """
        super(DequeOutLRUCache, self).__init__(maxsize, **kw)
        if out_deque is None:
            out_deque = collections.deque()
        elif not isinstance(out_deque, collections.deque):
            raise ValueError(u'out_deque should be collections.deque')
        self._out_deque = out_deque
        self._tracking = {}

    def __setitem__(self, key, value, **kw):
        super(DequeOutLRUCache, self).__setitem__(key, value, **kw)
        self._tracking[key] = value

    @property
    def out_deque(self):
        """The :class:`collections.deque` to which expired items are added."""
        expired = {k: v for (k, v) in self._tracking.items() if self.get(k) is None}
        for k, v in expired.items():
            del self._tracking[k]
            self._out_deque.append(v)
        return self._out_deque


class LockedObject(object):
    """LockedObject protects an object with a re-entrant lock.

    The lock is required by the context manager protocol.
    """
    # pylint: disable=too-few-public-methods

    def __init__(self, obj):
        self._lock = threading.RLock()
        self._obj = obj

    def __enter__(self):
        self._lock.acquire()
        return self._obj

    def __exit__(self, _exc_type, _exc_val, _exc_tb):
        self._lock.release()


def to_cache_timer(datetime_func):
    """Converts a datetime_func to a timestamp_func.

    Args:
       datetime_func (callable[[datatime]]): a func that returns the current
         time

    Returns:
       time_func (callable[[timestamp]): a func that returns the timestamp
         from the epoch
    """
    if datetime_func is None:
        datetime_func = datetime.utcnow

    def _timer():
        """Return the timestamp since the epoch."""
        return (datetime_func() - datetime(1970, 1, 1)).total_seconds()

    return _timer



# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""metric values provides funcs using to aggregate `MetricValue`.

:func:`merge` merges two `MetricValue` instances.
:func:`update_hash` adds a `MetricValue` to a secure hash
:func:`sign` generates a signature for a `MetricValue` using a secure hash

"""

from __future__ import absolute_import

import hashlib
import logging

from apitools.base.py import encoding

from . import distribution, money, signing, timestamp, MetricKind
from ..gen.servicecontrol_v1_messages import MetricValue


_logger = logging.getLogger(__name__)


def create(labels=None, **kw):
    """Constructs a new metric value.

    This acts as an alternate to MetricValue constructor which
    simplifies specification of labels.  Rather than having to create
    a MetricValue.Labels instance, all that's necessary to specify the
    required string.

    Args:
      labels (dict([string, [string]]):
      **kw: any other valid keyword args valid in the MetricValue constructor

    Returns
      :class:`MetricValue`: the created instance

    """
    if labels is not None:
        kw[u'labels'] = encoding.PyValueToMessage(MetricValue.LabelsValue,
                                                  labels)
    return MetricValue(**kw)


def merge(metric_kind, prior, latest):
    """Merges `prior` and `latest`

    Args:
       metric_kind (:class:`MetricKind`): indicates the kind of metrics
         being merged
       prior (:class:`MetricValue`): an prior instance of the metric
       latest (:class:`MetricValue`: the latest instance of the metric
    """
    prior_type, _ = _detect_value(prior)
    latest_type, _ = _detect_value(latest)
    if prior_type != latest_type:
        _logger.warn(u'Metric values are not compatible: %s, %s',
                     prior, latest)
        raise ValueError(u'Incompatible delta metric values')
    if prior_type is None:
        _logger.warn(u'Bad metric values, types not known for : %s, %s',
                     prior, latest)
        raise ValueError(u'Unsupported delta metric types')

    if metric_kind == MetricKind.DELTA:
        return _merge_delta_metric(prior, latest)
    else:
        return _merge_cumulative_or_gauge_metrics(prior, latest)


def update_hash(a_hash, mv):
    """Adds ``mv`` to ``a_hash``

    Args:
       a_hash (`Hash`): the secure hash, e.g created by hashlib.md5
       mv (:class:`MetricValue`): the instance to add to the hash

    """
    if mv.labels:
        signing.add_dict_to_hash(a_hash, encoding.MessageToPyValue(mv.labels))
    money_value = mv.get_assigned_value(u'moneyValue')
    if money_value is not None:
        a_hash.update(b'\x00')
        a_hash.update(money_value.currencyCode.encode('utf-8'))


def sign(mv):
    """Obtains a signature for a `MetricValue`

    Args:
       mv (:class:`endpoints_management.gen.servicecontrol_v1_messages.MetricValue`): a
         MetricValue that's part of an operation

    Returns:
       string: a unique signature for that operation
    """
    md5 = hashlib.md5()
    update_hash(md5, mv)
    return md5.digest()


def _merge_cumulative_or_gauge_metrics(prior, latest):
    if timestamp.compare(prior.endTime, latest.endTime) == -1:
        return latest
    else:
        return prior


def _merge_delta_metric(prior, latest):
    prior_type, prior_value = _detect_value(prior)
    latest_type, latest_value = _detect_value(latest)
    _merge_delta_timestamps(prior, latest)
    updated_value = _combine_delta_values(prior_type, prior_value, latest_value)
    setattr(latest, latest_type, updated_value)
    return latest


# This is derived from the oneof choices for the MetricValue message's value
# field in google/api/servicecontrol/v1/metric_value.proto, and should be kept
# in sync with that
_METRIC_VALUE_ONEOF_FIELDS = (
    u'boolValue', u'distributionValue', u'doubleValue', u'int64Value',
    u'moneyValue', u'stringValue')


def _detect_value(metric_value):
    for f in _METRIC_VALUE_ONEOF_FIELDS:
        value = metric_value.get_assigned_value(f)
        if value is not None:
            return f, value
    return None, None


def _merge_delta_timestamps(prior, latest):
    # Update the start time and end time in the latest metric value
    if (prior.startTime and
        (latest.startTime is None or
         timestamp.compare(prior.startTime, latest.startTime) == -1)):
        latest.startTime = prior.startTime

    if (prior.endTime and
        (latest.endTime is None or timestamp.compare(
            latest.endTime, prior.endTime) == -1)):
        latest.endTime = prior.endTime

    return latest


def _combine_delta_values(value_type, prior, latest):
    if value_type in (u'int64Value', u'doubleValue'):
        return prior + latest
    elif value_type == u'moneyValue':
        return money.add(prior, latest, allow_overflow=True)
    elif value_type == u'distributionValue':
        distribution.merge(prior, latest)
        return latest
    else:
        _logger.error(u'Unmergeable metric type %s', value_type)
        raise ValueError(u'Could not merge unmergeable metric type')



# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""wsgi implement behaviour that provides service control as wsgi
middleware.

It provides the :class:`Middleware`, which is a WSGI middleware implementation
that wraps another WSGI application to uses a provided
:class:`endpoints_management.control.client.Client` to provide service control.

"""
# pylint: disable=too-many-arguments

from __future__ import absolute_import

from datetime import datetime
import httplib
import logging
import os
import socket
import uuid
import urllib2
import urlparse
import wsgiref.util

import backoff
from webob.exc import HTTPServiceUnavailable, status_map as exc_status_map

from ..auth import suppliers, tokens
from ..config.service_config import ServiceConfigException
from . import check_request, client, quota_request, report_request, service, sm_messages


_logger = logging.getLogger(__name__)


_CONTENT_LENGTH = u'content-length'
_DEFAULT_LOCATION = u'global'

_METADATA_SERVER_URL = u'http://metadata.google.internal'


def _running_on_gce():
    headers = {u'Metadata-Flavor': u'Google'}

    try:
        request = urllib2.Request(_METADATA_SERVER_URL, headers=headers)
        response = urllib2.urlopen(request)
        if response.info().getheader(u'Metadata-Flavor') == u'Google':
            return True
    except (urllib2.URLError, socket.error):
        pass

    return False


def _get_platform():
    server_software = os.environ.get(u'SERVER_SOFTWARE', u'')

    if server_software.startswith(u'Development'):
        return report_request.ReportedPlatforms.DEVELOPMENT
    elif os.environ.get(u'KUBERNETES_SERVICE_HOST'):
        return report_request.ReportedPlatforms.GKE
    elif _running_on_gce():
        # We're either in GAE Flex or GCE
        if os.environ.get(u'GAE_MODULE_NAME'):
            return report_request.ReportedPlatforms.GAE_FLEX
        else:
            return report_request.ReportedPlatforms.GCE
    elif server_software.startswith(u'Google App Engine'):
        return report_request.ReportedPlatforms.GAE_STANDARD

    return report_request.ReportedPlatforms.UNKNOWN


platform = _get_platform()


def running_on_devserver():
    return platform == report_request.ReportedPlatforms.DEVELOPMENT


def add_all(application, project_id, control_client,
            loader=service.Loaders.FROM_SERVICE_MANAGEMENT):
    """Adds all endpoints middleware to a wsgi application.

    Sets up application to use all default endpoints middleware.

    Example:

      >>> application = MyWsgiApp()  # an existing WSGI application
      >>>
      >>> # the name of the controlled service
      >>> service_name = 'my-service-name'
      >>>
      >>> # A GCP project  with service control enabled
      >>> project_id = 'my-project-id'
      >>>
      >>> # wrap the app for service control
      >>> from endpoints_management.control import wsgi
      >>> control_client = client.Loaders.DEFAULT.load(service_name)
      >>> control_client.start()
      >>> wrapped_app = add_all(application, project_id, control_client)
      >>>
      >>> # now use wrapped_app in place of app

    Args:
       application: the wrapped wsgi application
       project_id: the project_id thats providing service control support
       control_client: the service control client instance
       loader (:class:`endpoints_management.control.service.Loader`): loads the service
          instance that configures this instance's behaviour
    """
    return ConfigFetchWrapper(application, project_id, control_client, loader)


class ConfigFetchWrapper(object):
    """
    This class encapsulates the service config loading process. If the initial
    loading attempt fails, it launches a background thread to retry with
    exponential backoff. However, if background threads are disabled, it will
    instead try loading the service config before every request.

    Since it might run in a situation where threading doesn't work, it does not
    use locks to coordinate access. Instead, the background thread may access
    only the self.loader and self.service_config variables; specifically,
    retrieving the former and retrieving and setting the latter. The Python GIL
    ensures that thread contexts can only switch between individual Python
    bytecodes.
    """
    def __init__(self, application, project_id, control_client,
                 loader=service.Loaders.FROM_SERVICE_MANAGEMENT,
                 disable_threading=False):
        self.service_config = None
        self.background_thread = None
        self.threading_failed = disable_threading
        # This will answer all requests with HTTP 503 Service Unavailable
        self.wsgi_backend = HTTPServiceUnavailable()

        self.application = application
        self.project_id = project_id
        self.control_client = control_client
        self.loader = loader

        self.try_loading()
        self.wrap_app()
        if self.service_config is None:
            self.launch_loading_thread()

    def __call__(self, environ, start_response):
        if self.threading_failed and self.service_config is None:
            self.try_loading()
        if isinstance(self.wsgi_backend, HTTPServiceUnavailable):
            self.wrap_app()
        return self.wsgi_backend(environ, start_response)

    def wrap_app(self):
        if self.service_config is None:
            return
        authenticator = _create_authenticator(self.service_config)

        wrapped_app = Middleware(self.application, self.project_id, self.control_client)
        if authenticator:
            wrapped_app = AuthenticationMiddleware(wrapped_app, authenticator)
        self.wsgi_backend = EnvironmentMiddleware(wrapped_app, self.service_config)

    def try_loading(self):
        try:
            a_service = self.loader.load()
            if not a_service:
                raise ValueError(u'Service config loader returned bad value.')
        except (ServiceConfigException, ValueError):
            _logger.exception(u'Failed to load service config.')
        else:
            _logger.debug('Loaded service config.')
            self.service_config = a_service

    def try_loading_in_thread(self):
        class LoadFailedException(Exception):
            pass

        @backoff.on_exception(backoff.expo, LoadFailedException)
        def _load_or_raise():
            self.try_loading()
            if self.service_config is None:
                raise LoadFailedException

        _load_or_raise()

    def launch_loading_thread(self):
        if self.threading_failed:
            return
        self.background_thread = client.create_thread(target=self.try_loading_in_thread)
        try:
            self.background_thread.start()
        except Exception:  # pylint: disable=broad-except
            _logger.exception(u'Failed to start service config loading background thread.')
            self.threading_failed = True
            self.background_thread = None

def _next_operation_uuid():
    return uuid.uuid4().hex


def _request_method(environ):
    return environ.get(u'HTTP_X_HTTP_METHOD_OVERRIDE', environ[u'REQUEST_METHOD'])


class EnvironmentMiddleware(object):
    """A WSGI middleware that sets related variables in the environment.

    It attempts to add the following vars:

    - google.api.config.service
    - google.api.config.service_name
    - google.api.config.method_registry
    - google.api.config.reporting_rules
    - google.api.config.method_info
    """
    # pylint: disable=too-few-public-methods

    SERVICE = u'google.api.config.service'
    SERVICE_NAME = u'google.api.config.service_name'
    METHOD_REGISTRY = u'google.api.config.method_registry'
    METHOD_INFO = u'google.api.config.method_info'
    REPORTING_RULES = u'google.api.config.reporting_rules'

    def __init__(self, application, a_service):
        """Initializes a new Middleware instance.

        Args:
          application: the wrapped wsgi application
          a_service (:class:`endpoints_management.gen.servicemanagement_v1_messages.Service`):
            a service instance
        """
        if not isinstance(a_service, sm_messages.Service):
            raise ValueError(u"service is None or not an instance of Service")

        self._application = application
        self._service = a_service

        method_registry, reporting_rules = self._configure()
        self._method_registry = method_registry
        self._reporting_rules = reporting_rules

    def _configure(self):
        registry = service.MethodRegistry(self._service)
        logs, metric_names, label_names = service.extract_report_spec(self._service)
        reporting_rules = report_request.ReportingRules.from_known_inputs(
            logs=logs,
            metric_names=metric_names,
            label_names=label_names)

        return registry, reporting_rules

    def __call__(self, environ, start_response):
        environ[self.SERVICE] = self._service
        environ[self.SERVICE_NAME] = self._service.name
        environ[self.METHOD_REGISTRY] = self._method_registry
        environ[self.REPORTING_RULES] = self._reporting_rules
        parsed_uri = urlparse.urlparse(wsgiref.util.request_uri(environ))
        http_method = _request_method(environ)
        method_info = self._method_registry.lookup(http_method, parsed_uri.path)
        if method_info:
            environ[self.METHOD_INFO] = method_info

        return self._application(environ, start_response)


class Middleware(object):
    """A WSGI middleware implementation that provides service control.

    Example:

      >>> app = MyWsgiApp()  # an existing WSGI application
      >>>
      >>> # the name of the controlled service
      >>> service_name = 'my-service-name'
      >>>
      >>> # A GCP project  with service control enabled
      >>> project_id = 'my-project-id'
      >>>
      >>> # wrap the app for service control
      >>> from endpoints_management.control import client, wsgi, service
      >>> control_client = client.Loaders.DEFAULT.load(service_name)
      >>> control_client.start()
      >>> wrapped_app = wsgi.Middleware(app, control_client, project_id)
      >>> env_app = wsgi.EnvironmentMiddleware(wrapped,app)
      >>>
      >>> # now use env_app in place of app

    """
    # pylint: disable=too-few-public-methods, fixme
    _NO_API_KEY_MSG = (
        u'Method does not allow callers without established identity.'
        u' Please use an API key or other form of API consumer identity'
        u' to call this API.')

    def __init__(self,
                 application,
                 project_id,
                 control_client,
                 next_operation_id=_next_operation_uuid,
                 timer=datetime.utcnow):
        """Initializes a new Middleware instance.

        Args:
           application: the wrapped wsgi application
           project_id: the project_id thats providing service control support
           control_client: the service control client instance
           next_operation_id (func): produces the next operation
           timer (func[[datetime.datetime]]): a func that obtains the current time
           """
        self._application = application
        self._project_id = project_id
        self._control_client = control_client
        self._next_operation_id = next_operation_id
        self._timer = timer

    def __call__(self, environ, start_response):
        # pylint: disable=too-many-locals
        method_info = environ.get(EnvironmentMiddleware.METHOD_INFO)
        if not method_info:
            # just allow the wrapped application to handle the request
            _logger.debug(u'method_info not present in the wsgi environment'
                          u', no service control')
            return self._application(environ, start_response)

        latency_timer = _LatencyTimer(self._timer)
        latency_timer.start()

        # Determine if the request can proceed
        http_method = _request_method(environ)
        parsed_uri = urlparse.urlparse(wsgiref.util.request_uri(environ))
        app_info = _AppInfo()
        # TODO: determine if any of the more complex ways of getting the request size
        # (e.g) buffering and counting the wsgi input stream is more appropriate here
        try:
            app_info.request_size = int(environ.get(u'CONTENT_LENGTH',
                                                    report_request.NOT_SET))
        except ValueError:
            _logger.warn(u'ignored bad content-length: %s', environ.get(u'CONTENT_LENGTH'))

        app_info.http_method = http_method
        app_info.url = parsed_uri

        # Default to 0 for consumer project number to disable per-consumer
        # metric reporting if the check request doesn't return one.
        consumer_project_number = 0
        check_info = self._create_check_info(method_info, parsed_uri, environ)
        if not check_info.api_key and not method_info.allow_unregistered_calls:
            _logger.debug(u"skipping %s, no api key was provided", parsed_uri)
            error_msg = self._handle_missing_api_key(app_info, start_response)
        else:
            check_req = check_info.as_check_request()
            _logger.debug(u'checking %s with %s', method_info, check_request)
            check_resp = self._control_client.check(check_req)
            error_msg = self._handle_check_response(app_info, check_resp, start_response)
            if (check_resp and check_resp.checkInfo and
                    check_resp.checkInfo.consumerInfo):
                consumer_project_number = (
                    check_resp.checkInfo.consumerInfo.projectNumber)
            if error_msg is None:
                quota_info = self._create_quota_info(method_info, parsed_uri, environ)
                if not quota_info.quota_info:
                    _logger.debug(u'no metric costs for this method')
                else:
                    quota_request = quota_info.as_allocate_quota_request()
                    quota_response = self._control_client.allocate_quota(quota_request)
                    error_msg = self._handle_quota_response(
                        app_info, quota_response, start_response)

        if error_msg:
            # send a report request that indicates that the request failed
            rules = environ.get(EnvironmentMiddleware.REPORTING_RULES)
            latency_timer.end()
            report_req = self._create_report_request(method_info,
                                                     check_info,
                                                     app_info,
                                                     latency_timer,
                                                     rules,
                                                     consumer_project_number)
            _logger.debug(u'scheduling report_request %s', report_req)
            self._control_client.report(report_req)
            return error_msg

        # update the client with the response
        latency_timer.app_start()

        # run the application request in an inner handler that sets the status
        # and response code on app_info
        def inner_start_response(status, response_headers, exc_info=None):
            app_info.response_code = int(status.partition(u' ')[0])
            for name, value in response_headers:
                if name.lower() == _CONTENT_LENGTH:
                    app_info.response_size = int(value)
                    break
            return start_response(status, response_headers, exc_info)

        result = self._application(environ, inner_start_response)

        # perform reporting, result must be joined otherwise the latency record
        # is incorrect
        result = b''.join(result)
        latency_timer.end()
        app_info.response_size = len(result)
        rules = environ.get(EnvironmentMiddleware.REPORTING_RULES)
        report_req = self._create_report_request(method_info,
                                                 check_info,
                                                 app_info,
                                                 latency_timer,
                                                 rules,
                                                 consumer_project_number)
        _logger.debug(u'scheduling report_request %s', report_req)
        self._control_client.report(report_req)
        return (result, )

    def _create_report_request(self,
                               method_info,
                               check_info,
                               app_info,
                               latency_timer,
                               reporting_rules,
                               consumer_project_number):
        # TODO: determine how to obtain the consumer_project_id and the location
        # correctly
        report_info = report_request.Info(
            api_key=check_info.api_key,
            api_key_valid=app_info.api_key_valid,
            api_method=method_info.selector,
            consumer_project_id=self._project_id,  # TODO: see above
            consumer_project_number=consumer_project_number,
            location=_DEFAULT_LOCATION,  # TODO: see above
            method=app_info.http_method,
            operation_id=check_info.operation_id,
            operation_name=check_info.operation_name,
            backend_time=latency_timer.backend_time,
            overhead_time=latency_timer.overhead_time,
            platform=platform,
            producer_project_id=self._project_id,
            protocol=report_request.ReportedProtocols.HTTP,
            request_size=app_info.request_size,
            request_time=latency_timer.request_time,
            response_code=app_info.response_code,
            response_size=app_info.response_size,
            referer=check_info.referer,
            service_name=check_info.service_name,
            url=app_info.url
        )
        return report_info.as_report_request(reporting_rules, timer=self._timer)

    def _get_api_key_info(self, method_info, parsed_uri, environ):
        api_key = _find_api_key_param(method_info, parsed_uri)
        if not api_key:
            api_key = _find_api_key_header(method_info, environ)
        if not api_key:
            api_key = _find_default_api_key_param(parsed_uri)
        return api_key

    def _create_check_info(self, method_info, parsed_uri, environ):
        service_name = environ.get(EnvironmentMiddleware.SERVICE_NAME)
        operation_id = self._next_operation_id()
        api_key = self._get_api_key_info(method_info, parsed_uri, environ)

        check_info = check_request.Info(
            android_cert_fingerprint=environ.get('HTTP_X_ANDROID_CERT', ''),
            android_package_name=environ.get('HTTP_X_ANDROID_PACKAGE', ''),
            api_key=api_key,
            api_key_valid=api_key is not None,
            client_ip=environ.get(u'REMOTE_ADDR', u''),
            consumer_project_id=self._project_id,  # TODO: switch this to producer_project_id
            ios_bundle_id=environ.get('HTTP_X_IOS_BUNDLE_IDENTIFIER', ''),
            operation_id=operation_id,
            operation_name=method_info.selector,
            referer=environ.get(u'HTTP_REFERER', u''),
            service_name=service_name
        )
        return check_info

    def _create_quota_info(self, method_info, parsed_uri, environ):
        service_name = environ.get(EnvironmentMiddleware.SERVICE_NAME)
        operation_id = self._next_operation_id()
        api_key = self._get_api_key_info(method_info, parsed_uri, environ)
        service = environ.get(EnvironmentMiddleware.SERVICE)

        return quota_request.Info(
            api_key=api_key,
            api_key_valid=api_key is not None,
            referer=environ.get(u'HTTP_REFERER', u''),
            consumer_project_id=self._project_id,
            operation_id=operation_id,
            operation_name=method_info.selector,
            service_name=service_name,
            quota_info=method_info.quota_info,
            config_id=service.id,
            client_ip=environ.get(u'REMOTE_ADDR', u''),
        )

    def _handle_check_response(self, app_info, check_resp, start_response):
        code, detail, api_key_valid = check_request.convert_response(
            check_resp, self._project_id)
        if code == httplib.OK:
            return None  # the check was OK

        # there was problem; the request cannot proceed
        _logger.warn(u'Check failed %d, %s', code, detail)
        app_info.response_code = code
        app_info.api_key_valid = api_key_valid
        return self._return_simple_http_response(start_response, code, detail)

    def _handle_quota_response(self, app_info, quota_resp, start_response):
        code, detail = quota_request.convert_response(
            quota_resp, self._project_id)
        if code == httplib.OK:
            return None  # the quota was OK

        # there was problem; the request cannot proceed
        _logger.warn(u'Quota failed %d, %s', code, detail)
        app_info.response_code = code
        return self._return_simple_http_response(start_response, code, detail)

    def _handle_missing_api_key(self, app_info, start_response):
        code = httplib.UNAUTHORIZED
        detail = self._NO_API_KEY_MSG
        _logger.warn(u'Check not performed %d, %s', code, detail)
        app_info.response_code = code
        app_info.api_key_valid = False
        return self._return_simple_http_response(start_response, code, detail)

    def _return_simple_http_response(self, start_response, code, detail):
        resp = exc_status_map[code](
            detail=detail, body_template = '''${explanation}\n\n${detail}\n''')
        # The resp (response) object is actually a very specialized
        # WSGI application, which means it accepts an environment
        # dictionary and a start_response callable. We don't have
        # access to the real request's WSGI environment at this point,
        # and even if we did, a lot of actual requests will simply
        # assume an application/json content type instead of
        # specifying it with the Accept header. So we just make up a
        # minimalistic WSGI environment; REQUEST_METHOD POST so that
        # it's fine to return a body in the response and Accept
        # application/json so the error response app will generate
        # JSON instead of HTML or text.
        return resp({'REQUEST_METHOD': 'POST', 'HTTP_ACCEPT': 'application/json'}, start_response)


class _AppInfo(object):
    # pylint: disable=too-few-public-methods

    def __init__(self):
        self.api_key_valid = True
        self.response_code = httplib.INTERNAL_SERVER_ERROR
        self.response_size = report_request.NOT_SET
        self.request_size = report_request.NOT_SET
        self.http_method = None
        self.url = None


class _LatencyTimer(object):

    def __init__(self, timer):
        self._timer = timer
        self._start = None
        self._app_start = None
        self._end = None

    def start(self):
        self._start = self._timer()

    def app_start(self):
        self._app_start = self._timer()

    def end(self):
        self._end = self._timer()
        if self._app_start is None:
            self._app_start = self._end

    @property
    def request_time(self):
        if self._start and self._end:
            return self._end - self._start
        return None

    @property
    def overhead_time(self):
        if self._start and self._app_start:
            return self._app_start - self._start
        return None

    @property
    def backend_time(self):
        if self._end and self._app_start:
            return self._end - self._app_start
        return None


def _find_api_key_param(info, parsed_uri):
    params = info.api_key_url_query_params
    if not params:
        return None

    param_dict = urlparse.parse_qs(parsed_uri.query)
    if not param_dict:
        return None

    for q in params:
        value = param_dict.get(q)
        if value:
            # param's values are lists, assume the first value
            # is what's needed
            return value[0]

    return None


_DEFAULT_API_KEYS = (u'key', u'api_key')


def _find_default_api_key_param(parsed_uri):
    param_dict = urlparse.parse_qs(parsed_uri.query)
    if not param_dict:
        return None

    for q in _DEFAULT_API_KEYS:
        value = param_dict.get(q)
        if value:
            # param's values are lists, assume the first value
            # is what's needed
            return value[0]

    return None


def _find_api_key_header(info, environ):
    headers = info.api_key_http_header
    if not headers:
        return None

    for h in headers:
        value = environ.get(u'HTTP_' + h.upper())
        if value:
            return value  # headers have single values

    return None


def _create_authenticator(a_service):
    """Create an instance of :class:`google.auth.tokens.Authenticator`.

    Args:
      a_service (:class:`endpoints_management.gen.servicemanagement_v1_messages.Service`): a
        service instance
    """
    if not isinstance(a_service, sm_messages.Service):
        raise ValueError(u"service is None or not an instance of Service")

    authentication = a_service.authentication
    if not authentication:
        _logger.info(u"authentication is not configured in service, "
                     u"authentication checks will be disabled")
        return

    issuers_to_provider_ids = {}
    issuer_uri_configs = {}
    for provider in authentication.providers:
        issuer = provider.issuer
        jwks_uri = provider.jwksUri

        # Enable openID discovery if jwks_uri is unset
        open_id = jwks_uri is None
        issuer_uri_configs[issuer] = suppliers.IssuerUriConfig(open_id, jwks_uri)
        issuers_to_provider_ids[issuer] = provider.id

    key_uri_supplier = suppliers.KeyUriSupplier(issuer_uri_configs)
    jwks_supplier = suppliers.JwksSupplier(key_uri_supplier)
    authenticator = tokens.Authenticator(issuers_to_provider_ids, jwks_supplier)
    return authenticator


class AuthenticationMiddleware(object):
    """A WSGI middleware that does authentication checks for incoming
    requests.

    In environments where os.environ is replaced with a request-local and
    thread-independent copy (e.g. Google Appengine), authentication result is
    added to os.environ so that the wrapped application can make use of the
    authentication result.
    """
    # pylint: disable=too-few-public-methods

    USER_INFO = u"google.api.auth.user_info"

    def __init__(self, application, authenticator):
        """Initializes an authentication middleware instance.

        Args:
          application: a WSGI application to be wrapped
          authenticator (:class:`google.auth.tokens.Authenticator`): an
            authenticator that authenticates incoming requests
        """
        if not isinstance(authenticator, tokens.Authenticator):
            raise ValueError(u"Invalid authenticator")

        self._application = application
        self._authenticator = authenticator

    def __call__(self, environ, start_response):
        method_info = environ.get(EnvironmentMiddleware.METHOD_INFO)
        if not method_info or not method_info.auth_info:
            # No authentication configuration for this method
            _logger.debug(u"authentication is not configured")
            return self._application(environ, start_response)

        auth_token = _extract_auth_token(environ)
        user_info = None
        if not auth_token:
            _logger.debug(u"No auth token is attached to the request")
        else:
            try:
                service_name = environ.get(EnvironmentMiddleware.SERVICE_NAME)
                user_info = self._authenticator.authenticate(auth_token,
                                                             method_info.auth_info,
                                                             service_name)
            except Exception:  # pylint: disable=broad-except
                _logger.debug(u"Cannot decode and verify the auth token. The backend "
                              u"will not be able to retrieve user info", exc_info=True)

        environ[self.USER_INFO] = user_info

        # pylint: disable=protected-access
        if user_info and not isinstance(os.environ, os._Environ):
            # Set user info into os.environ only if os.environ is replaced
            # with a request-local copy
            os.environ[self.USER_INFO] = user_info

        response = self._application(environ, start_response)

        # Erase user info from os.environ for safety and sanity.
        if self.USER_INFO in os.environ:
            del os.environ[self.USER_INFO]

        return response


_ACCESS_TOKEN_PARAM_NAME = u"access_token"
_BEARER_TOKEN_PREFIX = u"Bearer "
_BEARER_TOKEN_PREFIX_LEN = len(_BEARER_TOKEN_PREFIX)


def _extract_auth_token(environ):
    # First try to extract auth token from HTTP authorization header.
    auth_header = environ.get(u"HTTP_AUTHORIZATION")
    if auth_header:
        if auth_header.startswith(_BEARER_TOKEN_PREFIX):
            return auth_header[_BEARER_TOKEN_PREFIX_LEN:]
        return

    # Then try to read auth token from query.
    parameters = urlparse.parse_qs(environ.get(u"QUERY_STRING", u""))
    if _ACCESS_TOKEN_PARAM_NAME in parameters:
        auth_token, = parameters[_ACCESS_TOKEN_PARAM_NAME]
        return auth_token



# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides a method for fetching Service Configuration from Google Service
Management API."""

from __future__ import absolute_import

import logging
import json
import os
import urllib3

from apitools.base.py import encoding
from ..gen import servicemanagement_v1_messages as messages
from oauth2client import client
from urllib3.contrib import appengine


_logger = logging.getLogger(__name__)

_GOOGLE_API_SCOPE = u"https://www.googleapis.com/auth/cloud-platform"

_SERVICE_MGMT_URL_TEMPLATE = (u"https://servicemanagement.googleapis.com"
                              u"/v1/services/{}/configs/{}")

_SERVICE_NAME_ENV_KEY = u"ENDPOINTS_SERVICE_NAME"
_SERVICE_VERSION_ENV_KEY = u"ENDPOINTS_SERVICE_VERSION"


class ServiceConfigException(Exception):
    pass


def fetch_service_config(service_name=None, service_version=None):
    """Fetches the service config from Google Service Management API.

    Args:
      service_name: the service name. When this argument is unspecified, this
        method uses the value of the "SERVICE_NAME" environment variable as the
        service name, and raises ValueError if the environment variable is unset.
      service_version: the service version. When this argument is unspecified,
        this method uses the value of the "SERVICE_VERSION" environment variable
        as the service version, and raises ValueError if the environment variable
        is unset.

    Returns: the fetched service config JSON object.

    Raises:
      ValueError: when the service name/version is neither provided as an
        argument or set as an environment variable; or when the fetched service
        config fails validation.
      Exception: when the Google Service Management API returns non-200 response.
    """
    if not service_name:
        service_name = _get_env_var_or_raise(_SERVICE_NAME_ENV_KEY)
    if not service_version:
        service_version = _get_service_version(_SERVICE_VERSION_ENV_KEY,
                                               service_name)

    _logger.debug(u'Contacting Service Management API for service %s version %s',
                  service_name, service_version)
    response = _make_service_config_request(service_name, service_version)
    _logger.debug(u'obtained service json from the management api:\n%s', response.data)
    service = encoding.JsonToMessage(messages.Service, response.data)
    _validate_service_config(service, service_name, service_version)
    return service


def _get_access_token():
    credentials = client.GoogleCredentials.get_application_default()
    if credentials.create_scoped_required():
        credentials = credentials.create_scoped(_GOOGLE_API_SCOPE)
    return credentials.get_access_token().access_token


def _get_http_client():
    # don't use the AppEngineManager when sockets access is enabled
    # see https://urllib3.readthedocs.io/en/latest/reference/urllib3.contrib.html#module-urllib3.contrib.appengine
    if appengine.is_appengine_sandbox() and 'GAE_USE_SOCKETS_HTTPLIB' not in os.environ:
        return appengine.AppEngineManager()
    return urllib3.PoolManager()


def _get_env_var_or_raise(env_variable_name):
    if env_variable_name not in os.environ:
        message_template = u'The "{}" environment variable is not set'
        _log_and_raise(ValueError, message_template.format(env_variable_name))
    return os.environ[env_variable_name]


def _make_service_config_request(service_name, service_version=''):
    url = _SERVICE_MGMT_URL_TEMPLATE.format(service_name,
                                            service_version).rstrip('/')

    http_client = _get_http_client()
    headers = {u"Authorization": u"Bearer {}".format(_get_access_token())}
    response = http_client.request(u"GET", url, headers=headers)

    status_code = response.status
    if status_code == 403:
        message = (u"No service '{0}' found or permission denied. If this is a new "
                   u"Endpoints service, make sure you've deployed the "
                   u"service config using gcloud.").format(service_name)
        _log_and_raise(ServiceConfigException, message)
    elif status_code == 404:
        message = (u"The service '{0}' was found, but no service config was "
                   u"found for version '{1}'.").format(service_name, service_version)
        _log_and_raise(ServiceConfigException, message)
    elif status_code != 200:
        message_template = u"Fetching service config failed (status code {})"
        _log_and_raise(ServiceConfigException, message_template.format(status_code))

    return response


def _get_service_version(env_variable_name, service_name):
    service_version = os.environ.get(env_variable_name)

    if service_version:
        return service_version

    _logger.debug(u'Contacting Service Management API for service %s', service_name)
    response = _make_service_config_request(service_name)
    _logger.debug(u'obtained service config list from api: \n%s', response.data)

    services = encoding.JsonToMessage(messages.ListServiceConfigsResponse,
                                      response.data)

    try:
        _logger.debug(u"found latest service version of %s",
                      services.serviceConfigs[0].id)
        return services.serviceConfigs[0].id
    except:
        # catches IndexError if no versions or anything else that would
        # indicate a failed reading of the response.
        message_template = u"Couldn't retrieve service version from environment or server"
        _log_and_raise(ServiceConfigException, message_template)


def _validate_service_config(service, expected_service_name,
                             expected_service_version):
    service_name = service.name
    if not service_name:
        _log_and_raise(ValueError, u"No service name in the service config")
    if service_name != expected_service_name:
        message_template = u"Unexpected service name in service config: {}"
        _log_and_raise(ValueError, message_template.format(service_name))

    service_version = service.id
    if not service_version:
        _log_and_raise(ValueError, u"No service version in the service config")
    if service_version != expected_service_version:
        message_template = u"Unexpected service version in service config: {}"
        _log_and_raise(ValueError, message_template.format(service_version))


def _log_and_raise(exception_class, message):
    _logger.error(message)
    raise exception_class(message)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import httmock
import json
import mock
import os
import sys
import unittest

from apitools.base.py import encoding
from endpoints_management.config import service_config
from endpoints_management.control import sm_messages
from oauth2client import client

class ServiceConfigFetchTest(unittest.TestCase):

    _ACCESS_TOKEN = u"test_access_token"

    _SERVICE_NAME = u"test_service_name"
    _SERVICE_VERSION = u"test_service_version"
    _SERVICE_CONFIG_JSON = {
        u"name": _SERVICE_NAME,
        u"id": _SERVICE_VERSION
    }
    _SERVICE_CONFIG_LIST_JSON = {
        u"serviceConfigs": [{
            u"name": _SERVICE_NAME,
            u"title": _SERVICE_NAME,
            u"documentation": {},
            u"usage": {},
            u"id": _SERVICE_VERSION
        }]
    }

    _credentials = mock.MagicMock()
    _get_http_client = mock.MagicMock()

    def setUp(self):
        os.environ[u"ENDPOINTS_SERVICE_NAME"] = ServiceConfigFetchTest._SERVICE_NAME
        os.environ[u"ENDPOINTS_SERVICE_VERSION"] = ServiceConfigFetchTest._SERVICE_VERSION

        self._set_up_default_credential()

    def test_no_service_name(self):
        del os.environ[u"ENDPOINTS_SERVICE_NAME"]
        message = u'The "ENDPOINTS_SERVICE_NAME" environment variable is not set'
        with self.assertRaisesRegexp(ValueError, message):
            service_config.fetch_service_config()

    @mock.patch(u"endpoints_management.config.service_config.client.GoogleCredentials",
                _credentials)
    @mock.patch(u"endpoints_management.config.service_config._get_http_client", _get_http_client)
    def test_no_service_version(self):
        del os.environ[u"ENDPOINTS_SERVICE_VERSION"]

        list_mock_response = mock.MagicMock()
        list_mock_response.status = 200
        list_mock_response.data = json.dumps(ServiceConfigFetchTest._SERVICE_CONFIG_LIST_JSON)

        config_mock_response = mock.MagicMock()
        config_mock_response.status = 200
        config_mock_response.data = json.dumps(ServiceConfigFetchTest._SERVICE_CONFIG_JSON)

        mock_http_client = mock.MagicMock()
        mock_http_client.request.side_effect = [list_mock_response, config_mock_response]
        ServiceConfigFetchTest._get_http_client.return_value = mock_http_client

        service = encoding.JsonToMessage(sm_messages.Service,
                                         json.dumps(self._SERVICE_CONFIG_JSON))
        self.assertEqual(service, service_config.fetch_service_config())

        self.assertEqual(2, mock_http_client.request.call_count)

        headers={u"Authorization": u"Bearer " + ServiceConfigFetchTest._ACCESS_TOKEN}

        template = service_config._SERVICE_MGMT_URL_TEMPLATE
        url1 = template.format(ServiceConfigFetchTest._SERVICE_NAME, '').rstrip('/')
        url2 = template.format(ServiceConfigFetchTest._SERVICE_NAME,
                               ServiceConfigFetchTest._SERVICE_VERSION)

        call1 = mock.call(u"GET", url1, headers=headers)
        call2 = mock.call(u"GET", url2, headers=headers)

        mock_http_client.request.assert_has_calls([call1, call2])

    @mock.patch(u"endpoints_management.config.service_config.client.GoogleCredentials",
                _credentials)
    @mock.patch(u"endpoints_management.config.service_config._get_http_client", _get_http_client)
    def test_fetch_service_config(self):
        mock_response = mock.MagicMock()
        mock_response.status = 200
        mock_response.data = json.dumps(ServiceConfigFetchTest._SERVICE_CONFIG_JSON)
        mock_http_client = mock.MagicMock()
        mock_http_client.request.return_value = mock_response
        ServiceConfigFetchTest._get_http_client.return_value = mock_http_client

        service = encoding.JsonToMessage(sm_messages.Service,
                                         json.dumps(self._SERVICE_CONFIG_JSON))
        self.assertEqual(service, service_config.fetch_service_config())

        template = service_config._SERVICE_MGMT_URL_TEMPLATE
        url = template.format(ServiceConfigFetchTest._SERVICE_NAME,
                              ServiceConfigFetchTest._SERVICE_VERSION)
        headers={u"Authorization": u"Bearer " + ServiceConfigFetchTest._ACCESS_TOKEN}
        mock_http_client.request.assert_called_once_with(u"GET", url,
                                                         headers=headers)

    @mock.patch(u"endpoints_management.config.service_config.client.GoogleCredentials",
                _credentials)
    @mock.patch(u"endpoints_management.config.service_config._get_http_client", _get_http_client)
    def test_fetch_service_config_failed_forbidden(self):
        mock_response = mock.MagicMock()
        mock_response.status = 403
        mock_http_client = mock.MagicMock()
        mock_http_client.request.return_value = mock_response
        ServiceConfigFetchTest._get_http_client.return_value = mock_http_client
        expected_message = (u"No service 'test_service_name' found or permission denied. "
                            u"If this is a new Endpoints service, make sure you've deployed "
                            u"the service config using gcloud.")
        with self.assertRaisesRegexp(service_config.ServiceConfigException, expected_message):
            service_config.fetch_service_config()

    @mock.patch(u"endpoints_management.config.service_config.client.GoogleCredentials",
                _credentials)
    @mock.patch(u"endpoints_management.config.service_config._get_http_client", _get_http_client)
    def test_fetch_service_config_failed_not_found(self):
        mock_response = mock.MagicMock()
        mock_response.status = 404
        mock_http_client = mock.MagicMock()
        mock_http_client.request.return_value = mock_response
        ServiceConfigFetchTest._get_http_client.return_value = mock_http_client
        expected_message = (u"The service 'test_service_name' was found, but no service config "
                            u"was found for version 'test_service_version'.")
        with self.assertRaisesRegexp(service_config.ServiceConfigException, expected_message):
            service_config.fetch_service_config()

    @mock.patch(u"endpoints_management.config.service_config.client.GoogleCredentials",
                _credentials)
    @mock.patch(u"endpoints_management.config.service_config._get_http_client", _get_http_client)
    def test_fetch_service_config_with_wrong_service_name(self):
        mock_response = mock.MagicMock()
        mock_response.status = 200
        config = copy.deepcopy(ServiceConfigFetchTest._SERVICE_CONFIG_JSON)
        config[u"name"] = u"incorrect_service_name"
        mock_response.data = json.dumps(config)
        mock_http_client = mock.MagicMock()
        mock_http_client.request.return_value = mock_response
        ServiceConfigFetchTest._get_http_client.return_value = mock_http_client

        message = (u"Unexpected service name in service config: " +
                   config[u"name"])
        with self.assertRaisesRegexp(ValueError, message):
            service_config.fetch_service_config()

    @mock.patch(u"endpoints_management.config.service_config.client.GoogleCredentials",
                _credentials)
    @mock.patch(u"endpoints_management.config.service_config._get_http_client", _get_http_client)
    def test_fetch_service_config_with_wrong_service_version(self):
        mock_response = mock.MagicMock()
        mock_response.status = 200
        config = copy.deepcopy(ServiceConfigFetchTest._SERVICE_CONFIG_JSON)
        config[u"id"] = u"incorrect_service_version"
        mock_response.data = json.dumps(config)
        mock_http_client = mock.MagicMock()
        mock_http_client.request.return_value = mock_response
        ServiceConfigFetchTest._get_http_client.return_value = mock_http_client

        message = (u"Unexpected service version in service config: " +
                   config[u"id"])
        with self.assertRaisesRegexp(ValueError, message):
            service_config.fetch_service_config()

    def _set_up_default_credential(self):
        default_credential = mock.MagicMock()
        ServiceConfigFetchTest._credentials.get_application_default.return_value \
            = default_credential
        default_credential.create_scoped.return_value = default_credential
        token = ServiceConfigFetchTest._ACCESS_TOKEN
        access_token = client.AccessTokenInfo(access_token=token, expires_in=None)
        default_credential.get_access_token.return_value = access_token



# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""operation provides support for working with `Operation` instances.

:class:`~endpoints_management.gen.servicecontrol_v1_message.Operation` represents
information regarding an operation, and is a key constituent of
:class:`~endpoints_management.gen.servicecontrol_v1_message.CheckRequest` and
:class:`~endpoints_management.gen.servicecontrol_v1_message.ReportRequests.

The :class:`.Aggregator` support this.

"""

from __future__ import absolute_import

import collections
import logging
from datetime import datetime

from apitools.base.py import encoding

from . import metric_value, sc_messages, timestamp, MetricKind

_logger = logging.getLogger(__name__)


class Info(
        collections.namedtuple(
            u'Info', [
                u'android_cert_fingerprint',
                u'android_package_name',
                u'api_key',
                u'api_key_valid',
                u'consumer_project_id',
                u'ios_bundle_id',
                u'operation_id',
                u'operation_name',
                u'referer',
                u'service_name',
            ])):
    """Holds basic information about an api call.

    This class is one of several used to mediate between the raw service
    control api surface and python frameworks. Client code can construct
    operations using this surface

    Attributes:
        android_cert_fingerprint (string): the SHA-1 signing-certificate
          fingerprint of the calling app, used when the provided api_key is
          restricted to certain Android apps
        android_package_name (string): the package name of the calling app,
          used when the provided api_key is restricted to certain Android apps
        api_key (string): the api key
        api_key_valid (bool): it the request has a valid api key. By default
          it is true, it will only be set to false if the api key cannot
          be validated by the service controller
        consumer_project_id (string): the project id of the api consumer
        ios_bundle_id (string): the bundle identifier of the calling app,
          used when the provided api_key is restricted to certain iOS apps
        operation_id (string): identity of the operation, which must be unique
          within the scope of the service. Calls to report and check on the
          same operation should carry the same operation id
        operation_name (string): the fully-qualified name of the operation
        referer (string): the referer header, or if not present the origin
        service_name(string): the name of service

    """
    # pylint: disable=too-many-arguments

    def __new__(cls,
                android_cert_fingerprint=u'',
                android_package_name=u'',
                api_key=u'',
                api_key_valid=False,
                consumer_project_id=u'',
                ios_bundle_id=u'',
                operation_id=u'',
                operation_name=u'',
                referer=u'',
                service_name=u''):
        """Invokes the base constructor with default values."""
        return super(cls, Info).__new__(
            cls,
            android_cert_fingerprint,
            android_package_name,
            api_key,
            api_key_valid,
            consumer_project_id,
            ios_bundle_id,
            operation_id,
            operation_name,
            referer,
            service_name)

    def as_operation(self, timer=datetime.utcnow):
        """Makes an ``Operation`` from this instance.

        Returns:
          an ``Operation``

        """
        now = timer()
        op = sc_messages.Operation(
            endTime=timestamp.to_rfc3339(now),
            startTime=timestamp.to_rfc3339(now),
            importance=sc_messages.Operation.ImportanceValueValuesEnum.LOW)
        if self.operation_id:
            op.operationId = self.operation_id
        if self.operation_name:
            op.operationName = self.operation_name
        if self.api_key and self.api_key_valid:
            op.consumerId = u'api_key:' + self.api_key
        elif self.consumer_project_id:
            op.consumerId = u'project:' + self.consumer_project_id
        return op


class Aggregator(object):
    """Container that implements operation aggregation.

    Thread compatible.
    """
    DEFAULT_KIND = MetricKind.DELTA
    """Used when kinds are not specified, or are missing a metric name"""

    def __init__(self, initial_op, kinds=None):
        """Constructor.

        If kinds is not specifed, all operations will be merged assuming
        they are of Kind ``DEFAULT_KIND``

        Args:
           initial_op (
             :class:`endpoints_management.gen.servicecontrol_v1_messages.Operation`): the
               initial version of the operation
           kinds (dict[string,[string]]): specifies the metric kind for
              each metric name

        """
        assert isinstance(initial_op, sc_messages.Operation)
        if kinds is None:
            kinds = {}
        self._kinds = kinds
        self._metric_values_by_name_then_sign = collections.defaultdict(dict)
        our_op = encoding.CopyProtoMessage(initial_op)
        self._merge_metric_values(our_op)
        our_op.metricValueSets = []
        self._op = our_op

    def as_operation(self):
        """Obtains a single `Operation` representing this instances contents.

        Returns:
           :class:`endpoints_management.gen.servicecontrol_v1_messages.Operation`
        """
        result = encoding.CopyProtoMessage(self._op)
        names = sorted(self._metric_values_by_name_then_sign.keys())
        for name in names:
            mvs = self._metric_values_by_name_then_sign[name]
            result.metricValueSets.append(
                sc_messages.MetricValueSet(
                    metricName=name, metricValues=mvs.values()))
        return result

    def add(self, other_op):
        """Combines `other_op` with the operation held by this aggregator.

        N.B. It merges the operations log entries and metric values, but makes
        the assumption the operation is consistent.  It's the callers
        responsibility to ensure consistency

        Args:
           other_op (
             class:`endpoints_management.gen.servicecontrol_v1_messages.Operation`):
             an operation merge into this one

        """
        self._op.logEntries.extend(other_op.logEntries)
        self._merge_timestamps(other_op)
        self._merge_metric_values(other_op)

    def _merge_metric_values(self, other_op):
        for value_set in other_op.metricValueSets:
            name = value_set.metricName
            kind = self._kinds.get(name, self.DEFAULT_KIND)
            by_signature = self._metric_values_by_name_then_sign[name]
            for mv in value_set.metricValues:
                signature = metric_value.sign(mv)
                prior = by_signature.get(signature)
                if prior is not None:
                    metric_value.merge(kind, prior, mv)
                by_signature[signature] = mv

    def _merge_timestamps(self, other_op):
        # Update the start time and end time in self._op  as needed
        if (other_op.startTime and
            (self._op.startTime is None or
             timestamp.compare(other_op.startTime, self._op.startTime) == -1)):
            self._op.startTime = other_op.startTime

        if (other_op.endTime and
            (self._op.endTime is None or timestamp.compare(
                self._op.endTime, other_op.endTime) == -1)):
            self._op.endTime = other_op.endTime



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import datetime
import time
import unittest2
from operator import attrgetter
from expects import be_none, equal, expect, raise_error

from apitools.base.py import encoding

from endpoints_management.control import (caches, label_descriptor,
                                          metric_value, sc_messages,
                                          metric_descriptor, report_request,
                                          timestamp)


class TestReportingRules(unittest2.TestCase):
    subject_cls = report_request.ReportingRules
    WANTED_LABELS = (label_descriptor.KnownLabels.REFERER,)
    WANTED_METRICS = (metric_descriptor.KnownMetrics.CONSUMER_REQUEST_COUNT,)

    def test_should_construct_with_no_args(self):
        rules = self.subject_cls()
        expect(rules).not_to(be_none)
        expect(rules.logs).to(equal(set()))
        expect(rules.metrics).to(equal(tuple()))
        expect(rules.labels).to(equal(tuple()))

    def test_should_construct_with_ok_expected_args(self):
        rules = self.subject_cls(logs=[u'wanted_log'],
                                 metrics=self.WANTED_METRICS,
                                 labels=self.WANTED_LABELS)
        expect(rules).not_to(be_none)
        expect(rules.logs).to(equal(set([u'wanted_log'])))
        expect(rules.metrics).to(equal(self.WANTED_METRICS))
        expect(rules.labels).to(equal(self.WANTED_LABELS))

    def test_should_construct_with_alt_constructor(self):
        rules = self.subject_cls.from_known_inputs()
        expect(rules).not_to(be_none)
        expect(rules.logs).to(equal(set()))
        expect(rules.metrics).to(equal(tuple()))
        expect(rules.labels).to(equal(tuple()))

    def test_should_construct_with_alt_constructor_with_ok_args(self):
        logs = [u'wanted_log', u'wanted_log']
        label_names = [x.label_name for x in self.WANTED_LABELS]
        metric_names = [x.metric_name for x in self.WANTED_METRICS]
        rules = self.subject_cls.from_known_inputs(
            logs=logs,
            label_names=label_names,
            metric_names=metric_names
        )
        expect(rules).not_to(be_none)
        expect(rules.logs).to(equal(set([u'wanted_log'])))
        expect(rules.metrics).to(equal(self.WANTED_METRICS))
        expect(rules.labels).to(equal(self.WANTED_LABELS))


_TEST_CONSUMER_ID = u'testConsumerID'
_TEST_OP1_NAME = u'testOp1'
_TEST_OP2_NAME = u'testOp2'
_WANTED_USER_AGENT = label_descriptor.USER_AGENT
_START_OF_EPOCH = datetime.datetime.utcfromtimestamp(0)
_START_OF_EPOCH_TIMESTAMP = timestamp.to_rfc3339(_START_OF_EPOCH)
_TEST_SERVICE_NAME = u'a_service_name'
_TEST_SIZE=1
_TEST_LATENCY=datetime.timedelta(seconds=7)
_EXPECTED_OK_LOG_ENTRY = sc_messages.LogEntry(
    name = u'endpoints-log',
    severity = sc_messages.LogEntry.SeverityValueValuesEnum.INFO,
    structPayload=encoding.PyValueToMessage(
        sc_messages.LogEntry.StructPayloadValue, {
            u'http_response_code': 200,
            u'http_method': u'GET',
            u'request_latency_in_ms': 7000.0,
            u'timestamp': time.mktime(_START_OF_EPOCH.timetuple()),
            u'response_size': 1,
            u'request_size': 1,
            u'referer': u'a_referer',
        }),
    timestamp=_START_OF_EPOCH_TIMESTAMP
)
_EXPECTED_NOK_LOG_ENTRY = sc_messages.LogEntry(
    name = u'endpoints-log',
    severity = sc_messages.LogEntry.SeverityValueValuesEnum.ERROR,
    structPayload=encoding.PyValueToMessage(
        sc_messages.LogEntry.StructPayloadValue, {
            u'http_response_code': 404,
            u'http_method': u'GET',
            u'request_latency_in_ms': 7000.0,
            u'timestamp': time.mktime(_START_OF_EPOCH.timetuple()),
            u'response_size': 1,
            u'request_size': 1,
            u'referer': u'a_referer',
            u'error_cause': u'internal',
        }),
    timestamp=_START_OF_EPOCH_TIMESTAMP
)

_WANTED_USER_AGENT = label_descriptor.USER_AGENT
_WANTED_SERVICE_AGENT = label_descriptor.SERVICE_AGENT
_WANTED_PLATFORM = u'Unknown'

_EXPECTED_OK_METRIC = metric_descriptor.KnownMetrics.CONSUMER_REQUEST_COUNT
_EXPECTED_NOK_METRIC = metric_descriptor.KnownMetrics.CONSUMER_ERROR_COUNT
_ADD_LOG_TESTS = [
    (report_request.Info(
        operation_id=u'an_op_id',
        operation_name=u'an_op_name',
        method=u'GET',
        referer=u'a_referer',
        backend_time=_TEST_LATENCY,
        overhead_time=_TEST_LATENCY,
        request_time=_TEST_LATENCY,
        request_size=_TEST_SIZE,
        response_size=_TEST_SIZE,
        service_name=_TEST_SERVICE_NAME),
     sc_messages.Operation(
         importance=sc_messages.Operation.ImportanceValueValuesEnum.LOW,
         logEntries=[_EXPECTED_OK_LOG_ENTRY],
         operationId=u'an_op_id',
         operationName=u'an_op_name',
         startTime=_START_OF_EPOCH_TIMESTAMP,
         endTime=_START_OF_EPOCH_TIMESTAMP)
    ),
    (report_request.Info(
        response_code=404,
        operation_id=u'an_op_id',
        operation_name=u'an_op_name',
        method=u'GET',
        referer=u'a_referer',
        backend_time=_TEST_LATENCY,
        overhead_time=_TEST_LATENCY,
        request_time=_TEST_LATENCY,
        request_size=_TEST_SIZE,
        response_size=_TEST_SIZE,
        service_name=_TEST_SERVICE_NAME),
     sc_messages.Operation(
         importance=sc_messages.Operation.ImportanceValueValuesEnum.LOW,
         logEntries=[_EXPECTED_NOK_LOG_ENTRY],
         operationId=u'an_op_id',
         operationName=u'an_op_name',
         startTime=_START_OF_EPOCH_TIMESTAMP,
         endTime=_START_OF_EPOCH_TIMESTAMP)
    )
]

_TEST_API_KEY = u'test_key'
_ADD_METRICS_TESTS = [
    (report_request.Info(
        operation_id=u'an_op_id',
        operation_name=u'an_op_name',
        method=u'GET',
        referer=u'a_referer',
        backend_time=_TEST_LATENCY,
        overhead_time=_TEST_LATENCY,
        request_time=_TEST_LATENCY,
        request_size=_TEST_SIZE,
        response_size=_TEST_SIZE,
        service_name=_TEST_SERVICE_NAME,
        api_key=_TEST_API_KEY,
        api_key_valid=True),
     sc_messages.Operation(
         importance=sc_messages.Operation.ImportanceValueValuesEnum.LOW,
         logEntries=[],
         labels=encoding.PyValueToMessage(
             sc_messages.Operation.LabelsValue, {
                 u'servicecontrol.googleapis.com/service_agent':
                     _WANTED_SERVICE_AGENT,
                 u'servicecontrol.googleapis.com/user_agent':
                     _WANTED_USER_AGENT,
                 u'servicecontrol.googleapis.com/platform':
                     _WANTED_PLATFORM,
             }),
         metricValueSets = [
             sc_messages.MetricValueSet(
                 metricName=_EXPECTED_OK_METRIC.metric_name,
                 metricValues=[
                     metric_value.create(int64Value=1),
                 ]
             ),
         ],
         consumerId=u'api_key:' + _TEST_API_KEY,
         operationId=u'an_op_id',
         operationName=u'an_op_name',
         startTime=_START_OF_EPOCH_TIMESTAMP,
         endTime=_START_OF_EPOCH_TIMESTAMP)
    ),
    (report_request.Info(
        response_code=404,
        operation_id=u'an_op_id',
        operation_name=u'an_op_name',
        method=u'GET',
        referer=u'a_referer',
        backend_time=_TEST_LATENCY,
        overhead_time=_TEST_LATENCY,
        request_time=_TEST_LATENCY,
        request_size=_TEST_SIZE,
        response_size=_TEST_SIZE,
        service_name=_TEST_SERVICE_NAME,
        api_key=_TEST_API_KEY,
        api_key_valid=True),
     sc_messages.Operation(
         importance=sc_messages.Operation.ImportanceValueValuesEnum.LOW,
         logEntries=[],
         labels=encoding.PyValueToMessage(
             sc_messages.Operation.LabelsValue, {
                 u'servicecontrol.googleapis.com/service_agent':
                     _WANTED_SERVICE_AGENT,
                 u'servicecontrol.googleapis.com/user_agent':
                     _WANTED_USER_AGENT,
                 u'servicecontrol.googleapis.com/platform':
                     _WANTED_PLATFORM,
             }),
         metricValueSets = [
             sc_messages.MetricValueSet(
                 metricName=_EXPECTED_OK_METRIC.metric_name,
                 metricValues=[
                     metric_value.create(int64Value=1),
                 ]
             ),
             sc_messages.MetricValueSet(
                 metricName=_EXPECTED_NOK_METRIC.metric_name,
                 metricValues=[
                     metric_value.create(int64Value=1),
                 ]
             ),
         ],
         consumerId=u'api_key:' + _TEST_API_KEY,
         operationId=u'an_op_id',
         operationName=u'an_op_name',
         startTime=_START_OF_EPOCH_TIMESTAMP,
         endTime=_START_OF_EPOCH_TIMESTAMP)
    ),
]

_EXPECTED_OK_LABEL = label_descriptor.KnownLabels.REFERER
_ADD_LABELS_TESTS = [
    (report_request.Info(
        operation_id=u'an_op_id',
        operation_name=u'an_op_name',
        method=u'GET',
        referer=u'a_referer',
        service_name=_TEST_SERVICE_NAME),
     sc_messages.Operation(
         importance=sc_messages.Operation.ImportanceValueValuesEnum.LOW,
         labels=encoding.PyValueToMessage(
             sc_messages.Operation.LabelsValue, {
                 _EXPECTED_OK_LABEL.label_name: u'a_referer',
                 u'servicecontrol.googleapis.com/service_agent':
                     _WANTED_SERVICE_AGENT,
                 u'servicecontrol.googleapis.com/user_agent':
                     _WANTED_USER_AGENT,
                 u'servicecontrol.googleapis.com/platform':
                     _WANTED_PLATFORM,

             }),
         logEntries=[],
         operationId=u'an_op_id',
         operationName=u'an_op_name',
         startTime=_START_OF_EPOCH_TIMESTAMP,
         endTime=_START_OF_EPOCH_TIMESTAMP)
    ),
]

KEYGETTER = attrgetter(u'key')


class TestInfo(unittest2.TestCase):

    def test_should_construct_with_no_args(self):
        expect(report_request.Info()).not_to(be_none)

    def test_should_raise_if_constructed_with_a_bad_protocol(self):
        testf = lambda: report_request.Info(protocol=object())
        # not a report_request.ReportedProtocols
        expect(testf).to(raise_error(ValueError))

    def test_should_raise_if_constructed_with_a_bad_platform(self):
        testf = lambda: report_request.Info(platform=object())
        expect(testf).to(raise_error(ValueError))

    def test_should_raise_if_constructed_with_a_bad_request_size(self):
        testf = lambda: report_request.Info(request_size=object())
        expect(testf).to(raise_error(ValueError))
        testf = lambda: report_request.Info(request_size=-2)
        expect(testf).to(raise_error(ValueError))

    def test_should_raise_if_constructed_with_a_bad_response_size(self):
        testf = lambda: report_request.Info(response_size=object())
        expect(testf).to(raise_error(ValueError))
        testf = lambda: report_request.Info(response_size=-2)
        expect(testf).to(raise_error(ValueError))

    def test_should_raise_if_constructed_with_a_bad_backend_time(self):
        testf = lambda: report_request.Info(backend_time=object())
        expect(testf).to(raise_error(ValueError))

    def test_should_raise_if_constructed_with_a_bad_overhead_time(self):
        testf = lambda: report_request.Info(overhead_time=object())
        expect(testf).to(raise_error(ValueError))

    def test_should_raise_if_constructed_with_a_bad_request_time(self):
        testf = lambda: report_request.Info(request_time=object())
        expect(testf).to(raise_error(ValueError))

    def test_should_raise_if_constructed_with_a_bad_error_cause(self):
        testf = lambda: report_request.Info(error_cause=object())
        expect(testf).to(raise_error(ValueError))

    def test_should_fail_as_report_request_on_incomplete_info(self):
        timer = _DateTimeTimer()
        incomplete = report_request.Info()  # has no service_name
        rules = report_request.ReportingRules()
        testf = lambda: incomplete.as_report_request(rules, timer=timer)
        expect(testf).to(raise_error(ValueError))

    def test_should_add_expected_logs_as_report_request(self):
        timer = _DateTimeTimer()
        rules = report_request.ReportingRules(logs=[u'endpoints-log'])
        for info, want in _ADD_LOG_TESTS:
            got = info.as_report_request(rules, timer=timer)
            expect(got.serviceName).to(equal(_TEST_SERVICE_NAME))
            # compare the log entry in detail to avoid instability when
            # comparing the operations directly
            wantLogEntry = want.logEntries[0]
            gotLogEntry = got.reportRequest.operations[0].logEntries[0]
            expect(gotLogEntry.name).to(equal(wantLogEntry.name))
            expect(gotLogEntry.timestamp).to(equal(wantLogEntry.timestamp))
            print u'got timestamp', gotLogEntry.timestamp
            print u'want timestamp', wantLogEntry.timestamp
            expect(gotLogEntry.severity).to(equal(wantLogEntry.severity))
            gotStruct = encoding.MessageToPyValue(gotLogEntry.structPayload)
            print u'got struct', gotStruct
            wantStruct = encoding.MessageToPyValue(wantLogEntry.structPayload)
            print u'want struct', wantStruct
            expect(gotStruct).to(equal(wantStruct))

    def test_should_add_expected_metric_as_report_request(self):
        timer = _DateTimeTimer()
        rules = report_request.ReportingRules(metrics=[
            _EXPECTED_OK_METRIC, _EXPECTED_NOK_METRIC
        ])
        for info, want in _ADD_METRICS_TESTS:
            got = info.as_report_request(rules, timer=timer)
            # These additional properties have no well-defined order, so sort them.
            got.reportRequest.operations[0].labels.additionalProperties.sort(key=KEYGETTER)
            want.labels.additionalProperties.sort(key=KEYGETTER)
            expect(got.serviceName).to(equal(_TEST_SERVICE_NAME))
            expect(got.reportRequest.operations[0]).to(equal(want))

    def test_should_add_expected_label_as_report_request(self):
        timer = _DateTimeTimer()
        rules = report_request.ReportingRules(labels=[
            _EXPECTED_OK_LABEL
        ])
        for info, want in _ADD_LABELS_TESTS:
            got = info.as_report_request(rules, timer=timer)
            # These additional properties have no well-defined order, so sort them.
            got.reportRequest.operations[0].labels.additionalProperties.sort(key=KEYGETTER)
            want.labels.additionalProperties.sort(key=KEYGETTER)
            expect(got.serviceName).to(equal(_TEST_SERVICE_NAME))
            expect(got.reportRequest.operations[0]).to(equal(want))


class TestAggregatorReport(unittest2.TestCase):
    SERVICE_NAME = u'service.report'

    def setUp(self):
        self.timer = _DateTimeTimer()
        self.agg = report_request.Aggregator(
            self.SERVICE_NAME, caches.ReportOptions())

    def test_should_fail_if_req_is_bad(self):
        testf = lambda: self.agg.report(object())
        expect(testf).to(raise_error(ValueError))
        testf = lambda: self.agg.report(None)
        expect(testf).to(raise_error(ValueError))

    def test_should_fail_if_service_name_does_not_match(self):
        req = _make_test_request(self.SERVICE_NAME + u'-will-not-match')
        testf = lambda: self.agg.report(req)
        expect(testf).to(raise_error(ValueError))

    def test_should_fail_if_check_request_is_missing(self):
        req = sc_messages.ServicecontrolServicesReportRequest(
            serviceName=self.SERVICE_NAME)
        testf = lambda: self.agg.report(req)
        expect(testf).to(raise_error(ValueError))


class TestAggregatorTheCannotCache(unittest2.TestCase):
    SERVICE_NAME = u'service.no_cache'

    def setUp(self):
        # -ve num_entries means no cache is present
        self.agg = report_request.Aggregator(
            self.SERVICE_NAME,
            caches.ReportOptions(num_entries=-1))

    def test_should_not_cache_responses(self):
        req = _make_test_request(self.SERVICE_NAME)
        expect(self.agg.report(req)).to(be_none)

    def test_should_have_empty_flush_response(self):
        expect(len(self.agg.flush())).to(equal(0))

    def test_should_have_none_as_flush_interval(self):
        expect(self.agg.flush_interval).to(be_none)


class TestCachingAggregator(unittest2.TestCase):
    SERVICE_NAME = u'service.with_cache'

    def setUp(self):
        self.timer = _DateTimeTimer()
        self.flush_interval = datetime.timedelta(seconds=1)
        options = caches.ReportOptions(flush_interval=self.flush_interval)
        self.agg = report_request.Aggregator(
            self.SERVICE_NAME, options, timer=self.timer)

    def test_should_have_option_flush_interval_as_the_flush_interval(self):
        expect(self.agg.flush_interval).to(equal(self.flush_interval))

    def test_should_not_cache_requests_with_important_operations(self):
        req = _make_test_request(
            self.SERVICE_NAME,
            importance=sc_messages.Operation.ImportanceValueValuesEnum.HIGH)
        agg = self.agg
        expect(agg.report(req)).to(be_none)

    def test_should_cache_requests_and_return_cached_ok(self):
        req = _make_test_request(self.SERVICE_NAME, n=2, start=0)
        agg = self.agg
        expect(agg.report(req)).to(equal(report_request.Aggregator.CACHED_OK))

    def test_should_cache_requests_and_batch_them_on_flush(self):
        req1 = _make_test_request(self.SERVICE_NAME, n=2, start=0)
        req2 = _make_test_request(self.SERVICE_NAME, n=2, start=2)

        agg = self.agg
        expect(agg.report(req1)).to(equal(report_request.Aggregator.CACHED_OK))
        expect(agg.report(req2)).to(equal(report_request.Aggregator.CACHED_OK))
        # no immediate requests for flush
        flushed_reqs = agg.flush()
        expect(len(flushed_reqs)).to(equal(0))

        self.timer.tick() # time passes ...
        self.timer.tick() # ... and is now past the flush_interval
        flushed_reqs = agg.flush()
        expect(len(flushed_reqs)).to(equal(1))
        flushed_ops = flushed_reqs[0].reportRequest.operations
        expect(len(flushed_ops)).to(equal(4)) # number of ops in the req{1,2}

    def test_should_aggregate_operations_in_requests(self):
        n = 261 # arbitrary
        agg = self.agg
        for _ in range(n):
            # many requests, but only two ops
            req = _make_test_request(self.SERVICE_NAME, n=2, start=0)
            expect(agg.report(req)).to(
                equal(report_request.Aggregator.CACHED_OK))

        # time passes ...
        self.timer.tick()
        self.timer.tick() # ... and is now past the flush_interval
        flushed_reqs = agg.flush()
        expect(len(flushed_reqs)).to(equal(1))
        flushed_ops = flushed_reqs[0].reportRequest.operations
        expect(len(flushed_ops)).to(equal(2)) # many requests, but only two ops

    def test_may_clear_aggregated_operations(self):
        n = 261 # arbitrary
        agg = self.agg
        for i in range(n):
            # many requests, but only two ops
            req = _make_test_request(self.SERVICE_NAME, n=2, start=0)
            expect(agg.report(req)).to(
                equal(report_request.Aggregator.CACHED_OK))

        # time passes ...
        agg.clear()  # the aggregator is cleared
        self.timer.tick()
        self.timer.tick() # ... and is now past the flush_interval
        flushed_reqs = agg.flush()
        expect(len(flushed_reqs)).to(equal(0))  # but there is nothing


class _DateTimeTimer(object):
    def __init__(self, auto=False):
        self.auto = auto
        self.time = datetime.datetime.utcfromtimestamp(0)

    def __call__(self):
        if self.auto:
            self.tick()
        return self.time

    def tick(self):
        self.time += datetime.timedelta(seconds=1)


def _make_op_names(n, start=0):
    return (u'testOp%d' % (x,) for x in range(start, start + n))


def _make_test_request(service_name, importance=None, n=3, start=0):
    if importance is None:
        importance = sc_messages.Operation.ImportanceValueValuesEnum.LOW
    op_names = _make_op_names(n, start=start)
    ops = [sc_messages.Operation(consumerId=_TEST_CONSUMER_ID,
                              operationName=op_name,
                              importance=importance) for op_name in op_names]
    if ops:
        ops[0].labels = encoding.PyValueToMessage(
            sc_messages.Operation.LabelsValue, {
                u'key1': u'always add a label to the first op'})
    report_request = sc_messages.ReportRequest(operations=ops)
    return sc_messages.ServicecontrolServicesReportRequest(
        serviceName=service_name,
        reportRequest=report_request)



# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Decodes and verifies the signature of auth tokens."""

from __future__ import absolute_import

import datetime
import jwkest
import time

from dogpile import cache
from jwkest import jws
from jwkest import jwt

from . import suppliers

INT_TYPES = (int, long)


class Authenticator(object):  # pylint: disable=too-few-public-methods
    """Decodes and verifies the signature of auth tokens."""

    def __init__(self, issuers_to_provider_ids, jwks_supplier, cache_capacity=200):
        """Construct an instance of AuthTokenDecoder.

        Args:
          issuers_to_provider_ids: a dictionary mapping from issuers to provider
            IDs defined in the service configuration.
          jwks_supplier: an instance of JwksSupplier that supplies JWKS based on
            issuer.
          cache_capacity: the cache_capacity with default value of 200.
        """
        self._issuers_to_provider_ids = issuers_to_provider_ids
        self._jwks_supplier = jwks_supplier

        arguments = {u"capacity": cache_capacity}
        expiration_time = datetime.timedelta(minutes=5)
        self._cache = cache.make_region().configure(u"lru_cache",
                                                    arguments=arguments,
                                                    expiration_time=expiration_time)

    def authenticate(self, auth_token, auth_info, service_name):
        """Authenticates the current auth token.

        Args:
          auth_token: the auth token.
          auth_info: the auth configurations of the API method being called.
          service_name: the name of this service.

        Returns:
          A constructed UserInfo object representing the identity of the caller.

        Raises:
          UnauthenticatedException: When
            * the issuer is not allowed;
            * the audiences are not allowed;
            * the auth token has already expired.
        """
        try:
            jwt_claims = self.get_jwt_claims(auth_token)
        except Exception as error:
            raise suppliers.UnauthenticatedException(u"Cannot decode the auth token",
                                                     error)
        _check_jwt_claims(jwt_claims)

        user_info = UserInfo(jwt_claims)

        issuer = user_info.issuer
        if issuer not in self._issuers_to_provider_ids:
            raise suppliers.UnauthenticatedException(u"Unknown issuer: " + issuer)
        provider_id = self._issuers_to_provider_ids[issuer]

        if not auth_info.is_provider_allowed(provider_id):
            raise suppliers.UnauthenticatedException(u"The requested method does not "
                                                     u"allow provider id: " + provider_id)

        # Check the audiences decoded from the auth token. The auth token is
        # allowed when 1) an audience is equal to the service name, or 2) at least
        # one audience is allowed in the method configuration.
        audiences = user_info.audiences
        has_service_name = service_name in audiences

        allowed_audiences = auth_info.get_allowed_audiences(provider_id)
        intersected_audiences = set(allowed_audiences).intersection(audiences)
        if not has_service_name and not intersected_audiences:
            raise suppliers.UnauthenticatedException(u"Audiences not allowed")

        return user_info

    def get_jwt_claims(self, auth_token):
        """Decodes the auth_token into JWT claims represented as a JSON object.

        This method first tries to look up the cache and returns the result
        immediately in case of a cache hit. When cache misses, the method tries to
        decode the given auth token, verify its signature, and check the existence
        of required JWT claims. When successful, the decoded JWT claims are loaded
        into the cache and then returned.

        Args:
          auth_token: the auth token to be decoded.

        Returns:
          The decoded JWT claims.

        Raises:
          UnauthenticatedException: When the signature verification fails, or when
            required claims are missing.
        """

        def _decode_and_verify():
            jwt_claims = jwt.JWT().unpack(auth_token).payload()
            _verify_required_claims_exist(jwt_claims)

            issuer = jwt_claims[u"iss"]
            keys = self._jwks_supplier.supply(issuer)
            try:
                return jws.JWS().verify_compact(auth_token, keys)
            except (jwkest.BadSignature, jws.NoSuitableSigningKeys,
                    jws.SignerAlgError) as exception:
                raise suppliers.UnauthenticatedException(u"Signature verification failed",
                                                         exception)

        return self._cache.get_or_create(auth_token, _decode_and_verify)


class UserInfo(object):
    """An object that holds the authentication results."""

    def __init__(self, jwt_claims):
        audiences = jwt_claims[u"aud"]
        if isinstance(audiences, basestring):
            audiences = [audiences]
        self._audiences = audiences

        # email is not required
        self._email = jwt_claims[u"email"] if u"email" in jwt_claims else None
        self._subject_id = jwt_claims[u"sub"]
        self._issuer = jwt_claims[u"iss"]

    @property
    def audiences(self):
        return self._audiences

    @property
    def email(self):
        return self._email

    @property
    def subject_id(self):
        return self._subject_id

    @property
    def issuer(self):
        return self._issuer


def _check_jwt_claims(jwt_claims):
    """Checks whether the JWT claims should be accepted.

    Specifically, this method checks the "exp" claim and the "nbf" claim (if
    present), and raises UnauthenticatedException if 1) the current time is
    before the time identified by the "nbf" claim, or 2) the current time is
    equal to or after the time identified by the "exp" claim.

    Args:
      jwt_claims: the JWT claims whose expiratio to be checked.

    Raises:
      UnauthenticatedException: When the "exp" claim is malformed or the JWT has
        already expired.
    """
    current_time = time.time()

    expiration = jwt_claims[u"exp"]
    if not isinstance(expiration, INT_TYPES):
        raise suppliers.UnauthenticatedException(u'Malformed claim: "exp" must be an integer')
    if current_time >= expiration:
        raise suppliers.UnauthenticatedException(u"The auth token has already expired")

    if u"nbf" not in jwt_claims:
        return

    not_before_time = jwt_claims[u"nbf"]
    if not isinstance(not_before_time, INT_TYPES):
        raise suppliers.UnauthenticatedException(u'Malformed claim: "nbf" must be an integer')
    if current_time < not_before_time:
        raise suppliers.UnauthenticatedException(u'Current time is less than the "nbf" time')


def _verify_required_claims_exist(jwt_claims):
    """Verifies that the required claims exist.

    Args:
      jwt_claims: the JWT claims to be verified.

    Raises:
      UnauthenticatedException: if some claim doesn't exist.
    """
    for claim_name in [u"aud", u"exp", u"iss", u"sub"]:
        if claim_name not in jwt_claims:
            raise suppliers.UnauthenticatedException(u'Missing "%s" claim' % claim_name)



# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""service provides funcs for working with ``Service`` instances.

:func:`extract_report_spec` obtains objects used to determine what metrics,
labels and logs are included in a report request.

:class:`MethodRegistry` obtains a registry of `MethodInfo` instances from the
data within a `Service` which can then be used to determine which methods get
tracked.

:class:`Loaders` enumerates the different ways in which to obtain a usable
``Service`` instance

"""

from __future__ import absolute_import

import collections
import logging
import os
import urllib


from apitools.base.py import encoding
from enum import Enum

from ..config import service_config
from . import label_descriptor, metric_descriptor, path_regex, sm_messages


_logger = logging.getLogger(__name__)


CONFIG_VAR = u'ENDPOINTS_SERVICE_CONFIG_FILE'


def _load_from_well_known_env():
    if CONFIG_VAR not in os.environ:
        _logger.warn(u'did not load service; no environ var %s', CONFIG_VAR)
        return None
    config_file = os.environ[CONFIG_VAR]
    if not os.path.exists(os.environ[CONFIG_VAR]):
        _logger.warn(u'did not load service; missing config file %s', config_file)
        return None
    try:
        with open(config_file) as f:
            return encoding.JsonToMessage(sm_messages.Service, f.read())
    except ValueError:
        _logger.warn(u'did not load service; bad json config file %s', config_file)
        return None


_SIMPLE_CONFIG = """
{
    "name": "allow-all",
    "http": {
        "rules": [{
            "selector": "allow-all.GET",
            "get": "{x}"
        }, {
            "selector": "allow-all.PATCH",
            "patch": "{x}"
        }, {
            "selector": "allow-all.POST",
            "post": "{x}"
        }]
    },
    "usage": {
        "rules": [{
            "selector" : "allow-all.GET",
            "allowUnregisteredCalls" : true
        }, {
            "selector" : "allow-all.PATCH",
            "allowUnregisteredCalls" : true
        }, {
            "selector" : "allow-all.POST",
            "allowUnregisteredCalls" : true
        }]
    }
}
"""
_SIMPLE_CORE = encoding.JsonToMessage(sm_messages.Service, _SIMPLE_CONFIG)


def _load_simple():
    return encoding.CopyProtoMessage(_SIMPLE_CORE)


class Loaders(Enum):
    """Enumerates the functions used to load service configs."""
    # pylint: disable=too-few-public-methods
    ENVIRONMENT = (_load_from_well_known_env,)
    SIMPLE = (_load_simple,)
    FROM_SERVICE_MANAGEMENT = (service_config.fetch_service_config,)

    def __init__(self, load_func):
        """Constructor.

        load_func is used to load a service config
        """
        self._load_func = load_func

    def load(self, **kw):
        return self._load_func(**kw)


class MethodRegistry(object):
    """Provides a registry of the api methods defined by a ``Service``.

    During construction, ``MethodInfo`` instances are extracted from a
    ``Service``.  The are subsequently accessible via the :func:`lookup` method.

    """
    # pylint: disable=too-few-public-methods
    _OPTIONS = u'OPTIONS'

    def __init__(self, service):
        """Constructor.

        Args:
          service (:class:`endpoints_management.gen.servicemanagement_v1_messages.Service`):
            a service instance
        """
        if not isinstance(service, sm_messages.Service):
            raise ValueError(u'service should be an instance of Service')
        if not service.name:
            raise ValueError(u'Bad service: the name is missing')

        self._service = service  # the service that provides the methods
        self._extracted_methods = {}  # tracks all extracted_methods by selector

        self._auth_infos = self._extract_auth_config()
        self._quota_infos = self._extract_quota_config()

        # tracks urls templates
        self._templates_method_infos = collections.defaultdict(list)
        self._extract_methods()

    def lookup(self, http_method, path):
        http_method = http_method.lower()
        if path.startswith(u'/'):
            path = path[1:]
        tmi = self._templates_method_infos.get(http_method)
        if not tmi:
            _logger.debug(u'No methods for http method %s in %s',
                          http_method,
                          self._templates_method_infos.keys())
            return None
        # need to remove url quoting of colons. this is the simplest way.
        path = path.replace('%3A', ':')
        # pylint: disable=fixme
        # TODO: speed this up if it proves to be bottleneck.
        #
        # There is sophisticated trie-based solution in esp, something similar
        # could be built around the path_template implementation
        for template, method_info in tmi:
            _logger.debug(u'trying %s with template %s', path, template.pattern)
            match = template.match(path)
            if match:
                _logger.debug(u'%s matched template %s', path, template.pattern)
                return method_info
            else:
                _logger.debug(u'%s did not match template %s', path, template.pattern)

        return None

    def _extract_auth_config(self):
        """Obtains the authentication configurations."""

        service = self._service
        if not service.authentication:
            return {}

        auth_infos = {}
        for auth_rule in service.authentication.rules:
            selector = auth_rule.selector
            provider_ids_to_audiences = {}
            for requirement in auth_rule.requirements:
                provider_id = requirement.providerId
                if provider_id and requirement.audiences:
                    audiences = requirement.audiences.split(u",")
                    provider_ids_to_audiences[provider_id] = audiences
            auth_infos[selector] = AuthInfo(provider_ids_to_audiences)
        return auth_infos

    def _extract_quota_config(self):
        service = self._service
        if not service.quota:
            return {}

        quota_infos = {}
        for metric_rule in service.quota.metricRules:
            selector = metric_rule.selector
            costs = {ap.key: ap.value for ap in metric_rule.metricCosts.additionalProperties}
            quota_infos[selector] = costs

        return quota_infos

    def _extract_methods(self):
        """Obtains the methods used in the service."""
        service = self._service
        all_urls = set()
        urls_with_options = set()
        if not service.http:
            return
        for rule in service.http.rules:
            http_method, url = _detect_pattern_option(rule)
            if not url or not http_method or not rule.selector:
                _logger.error(u'invalid HTTP binding encountered')
                continue

            # Obtain the method info
            method_info = self._get_or_create_method_info(rule.selector)
            if rule.body:
                method_info.body_field_path = rule.body
            if not self._register(http_method, url, method_info):
                continue  # detected an invalid url
            all_urls.add(url)

            if http_method == self._OPTIONS:
                urls_with_options.add(url)

        self._add_cors_options_selectors(all_urls - urls_with_options)
        self._update_usage()
        self._update_system_parameters()

    def _register(self, http_method, url, method_info):
        if url.startswith('/'):
            url = url[1:]
        try:
            http_method = http_method.lower()
            template = path_regex.compile_path_pattern(url)
            self._templates_method_infos[http_method].append((template, method_info))
            _logger.debug(u'Registered template %s under method %s',
                          template.pattern,
                          http_method)
            return True
        except path_regex.RegexError:
            _logger.error(u'invalid HTTP template provided: %s', url)
            return False

    def _update_usage(self):
        extracted_methods = self._extracted_methods
        service = self._service
        if not service.usage:
            return
        for rule in service.usage.rules:
            selector = rule.selector
            method = extracted_methods.get(selector)
            if method:
                method.allow_unregistered_calls = rule.allowUnregisteredCalls
            else:
                _logger.error(u'bad usage selector: No HTTP rule for %s', selector)

    def _get_or_create_method_info(self, selector):
        extracted_methods = self._extracted_methods
        info = self._extracted_methods.get(selector)
        if info:
            return info

        auth_infos = self._auth_infos
        quota_infos = self._quota_infos
        auth_info = auth_infos[selector] if selector in auth_infos else None
        quota_info = quota_infos[selector] if selector in quota_infos else None

        info = MethodInfo(selector, auth_info, quota_info)
        extracted_methods[selector] = info
        return info

    def _add_cors_options_selectors(self, urls):
        extracted_methods = self._extracted_methods
        base_selector = u'%s.%s' % (self._service.name, self._OPTIONS)

        # ensure that no existing options selector is being used
        options_selector = base_selector
        n = 0
        while extracted_methods.get(options_selector) is not None:
            n += 1
            options_selector = u'%s.%d' % (base_selector, n)
        method_info = self._get_or_create_method_info(options_selector)
        method_info.allow_unregistered_calls = True
        for u in urls:
            self._register(self._OPTIONS, u, method_info)

    def _update_system_parameters(self):
        extracted_methods = self._extracted_methods
        service = self._service
        if not service.systemParameters:
            return
        rules = service.systemParameters.rules
        for rule in rules:
            selector = rule.selector
            method = extracted_methods.get(selector)
            if not method:
                _logger.error(u'bad system parameter: No HTTP rule for %s',
                              selector)
                continue

            for parameter in rule.parameters:
                name = parameter.name
                if not name:
                    _logger.error(u'bad system parameter: no parameter name %s',
                                  selector)
                    continue

                if parameter.httpHeader:
                    method.add_header_param(name, parameter.httpHeader)
                if parameter.urlQueryParameter:
                    method.add_url_query_param(name, parameter.urlQueryParameter)


class AuthInfo(object):
    """Consolidates auth information about methods defined in a ``Service``."""

    def __init__(self, provider_ids_to_audiences):
        """Construct an AuthInfo instance.

        Args:
          provider_ids_to_audiences: a dictionary that maps from provider ids
            to allowed audiences.
        """
        self._provider_ids_to_audiences = provider_ids_to_audiences

    def is_provider_allowed(self, provider_id):
        return provider_id in self._provider_ids_to_audiences

    def get_allowed_audiences(self, provider_id):
        return self._provider_ids_to_audiences.get(provider_id, [])


class MethodInfo(object):
    """Consolidates information about methods defined in a ``Service``."""
    API_KEY_NAME = u'api_key'
    # pylint: disable=too-many-instance-attributes

    def __init__(self, selector, auth_info, quota_info):
        self.selector = selector
        self.auth_info = auth_info
        self.quota_info = quota_info
        self.allow_unregistered_calls = False
        self.backend_address = u''
        self.body_field_path = u''
        self._url_query_parameters = collections.defaultdict(list)
        self._header_parameters = collections.defaultdict(list)

    def add_url_query_param(self, name, parameter):
        self._url_query_parameters[name].append(parameter)

    def add_header_param(self, name, parameter):
        self._header_parameters[name].append(parameter)

    def url_query_param(self, name):
        return tuple(self._url_query_parameters[name])

    def header_param(self, name):
        return tuple(self._header_parameters[name])

    @property
    def api_key_http_header(self):
        return self.header_param(self.API_KEY_NAME)

    @property
    def api_key_url_query_params(self):
        return self.url_query_param(self.API_KEY_NAME)


def extract_report_spec(
        service,
        label_is_supported=label_descriptor.KnownLabels.is_supported,
        metric_is_supported=metric_descriptor.KnownMetrics.is_supported):
    """Obtains the used logs, metrics and labels from a service.

    label_is_supported and metric_is_supported are filter functions used to
    determine if label_descriptors or metric_descriptors found in the service
    are supported.

    Args:
       service (:class:`endpoints_management.gen.servicecontrol_v1_messages.Service`):
          a service instance
       label_is_supported (:func): determines if a given label is supported
       metric_is_supported (:func): determines if a given metric is supported

    Return:
       tuple: (
         logs (set[string}), # the logs to report to
         metrics (list[string]), # the metrics to use
         labels (list[string]) # the labels to add
       )
    """
    resource_descs = service.monitoredResources
    labels_dict = {}
    logs = set()
    if service.logging:
        logs = _add_logging_destinations(
            service.logging.producerDestinations,
            resource_descs,
            service.logs,
            labels_dict,
            label_is_supported
        )
    metrics_dict = {}
    monitoring = service.monitoring
    if monitoring:
        for destinations in (monitoring.consumerDestinations,
                             monitoring.producerDestinations):
            _add_monitoring_destinations(destinations,
                                         resource_descs,
                                         service.metrics,
                                         metrics_dict,
                                         metric_is_supported,
                                         labels_dict,
                                         label_is_supported)
    return logs, metrics_dict.keys(), labels_dict.keys()


def _add_logging_destinations(destinations,
                              resource_descs,
                              log_descs,
                              labels_dict,
                              is_supported):
    all_logs = set()
    for d in destinations:
        if not _add_labels_for_a_monitored_resource(resource_descs,
                                                    d.monitoredResource,
                                                    labels_dict,
                                                    is_supported):
            continue  # skip bad monitored resources
        for log in d.logs:
            if _add_labels_for_a_log(log_descs, log, labels_dict, is_supported):
                all_logs.add(log)  # only add correctly configured logs
    return all_logs


def _add_monitoring_destinations(destinations,
                                 resource_descs,
                                 metric_descs,
                                 metrics_dict,
                                 metric_is_supported,
                                 labels_dict,
                                 label_is_supported):
    # pylint: disable=too-many-arguments
    for d in destinations:
        if not _add_labels_for_a_monitored_resource(resource_descs,
                                                    d.monitoredResource,
                                                    labels_dict,
                                                    label_is_supported):
            continue  # skip bad monitored resources
        for metric_name in d.metrics:
            metric_desc = _find_metric_descriptor(metric_descs, metric_name,
                                                  metric_is_supported)
            if not metric_desc:
                continue  # skip unrecognized or unsupported metric
            if not _add_labels_from_descriptors(metric_desc.labels, labels_dict,
                                                label_is_supported):
                continue  # skip metrics with bad labels
            metrics_dict[metric_name] = metric_desc


def _add_labels_from_descriptors(descs, labels_dict, is_supported):
    # only add labels if there are no conflicts
    for desc in descs:
        existing = labels_dict.get(desc.key)
        if existing and existing.valueType != desc.valueType:
            _logger.warn(u'halted label scan: conflicting label in %s', desc.key)
            return False
    # Update labels_dict
    for desc in descs:
        if is_supported(desc):
            labels_dict[desc.key] = desc
    return True


def _add_labels_for_a_log(logging_descs, log_name, labels_dict, is_supported):
    for d in logging_descs:
        if d.name == log_name:
            _add_labels_from_descriptors(d.labels, labels_dict, is_supported)
            return True
    _logger.warn(u'bad log label scan: log not found %s', log_name)
    return False


def _add_labels_for_a_monitored_resource(resource_descs,
                                         resource_name,
                                         labels_dict,
                                         is_supported):
    for d in resource_descs:
        if d.type == resource_name:
            _add_labels_from_descriptors(d.labels, labels_dict, is_supported)
            return True
    _logger.warn(u'bad monitored resource label scan: resource not found %s',
                resource_name)
    return False


def _find_metric_descriptor(metric_descs, name, metric_is_supported):
    for d in metric_descs:
        if name != d.name:
            continue
        if metric_is_supported(d):
            return d
        else:
            return None
    return None


# This is derived from the oneof choices of the HttpRule message's pattern
# field in google/api/http.proto, and should be kept in sync with that
_HTTP_RULE_ONE_OF_FIELDS = (
    u'get', u'put', u'post', u'delete', u'patch', u'custom')


def _detect_pattern_option(http_rule):
    for f in _HTTP_RULE_ONE_OF_FIELDS:
        value = http_rule.get_assigned_value(f)
        if value is not None:
            if f == u'custom':
                return value.kind, value.path
            else:
                return f, value
    return None, None



# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements a utility for parsing path templates."""

# This is ported over from endpoints.api_config_manager.

from __future__ import absolute_import

import base64
import re

# Internal constants
_PATH_VARIABLE_PATTERN = r'[a-zA-Z_][a-zA-Z_.\d]*'
_PATH_VALUE_PATTERN = r'[^/?#\[\]{}]*'

RegexError = re.error  # convenient alias


def _to_safe_path_param_name(matched_parameter):
    """Creates a safe string to be used as a regex group name.

    Only alphanumeric characters and underscore are allowed in variable name
    tokens, and numeric are not allowed as the first character.

    We cast the matched_parameter to base32 (since the alphabet is safe),
    strip the padding (= not safe) and prepend with _, since we know a token
    can begin with underscore.

    Args:
      matched_parameter: A string containing the parameter matched from the URL
        template.

    Returns:
      A string that's safe to be used as a regex group name.
    """
    return '_' + base64.b32encode(matched_parameter).rstrip('=')

def compile_path_pattern(pattern):
    r"""Generates a compiled regex pattern for a path pattern.

    e.g. '/MyApi/v1/notes/{id}'
    returns re.compile(r'/MyApi/v1/notes/(?P<id>[^/?#\[\]{}]*)')

    Args:
      pattern: A string, the parameterized path pattern to be checked.

    Returns:
      A compiled regex object to match this path pattern.
    """

    def replace_variable(match):
      """Replaces a {variable} with a regex to match it by name.

      Changes the string corresponding to the variable name to the base32
      representation of the string, prepended by an underscore. This is
      necessary because we can have message variable names in URL patterns
      (e.g. via {x.y}) but the character '.' can't be in a regex group name.

      Args:
        match: A regex match object, the matching regex group as sent by
          re.sub().

      Returns:
        A string regex to match the variable by name, if the full pattern was
        matched.
      """
      if match.lastindex > 1:
        var_name = _to_safe_path_param_name(match.group(2))
        return '%s(?P<%s>%s)' % (match.group(1), var_name,
                                 _PATH_VALUE_PATTERN)
      return match.group(0)

    pattern = re.sub('(/|^){(%s)}(?=/|$|:)' % _PATH_VARIABLE_PATTERN,
                     replace_variable, pattern)
    return re.compile(pattern + '/?$')



# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines several suppliers that are used by the authenticator."""

from __future__ import absolute_import

import datetime
from dogpile import cache
from jwkest import jwk
import requests
import ssl


_HTTP_PROTOCOL_PREFIX = u"http://"
_HTTPS_PROTOCOL_PREFIX = u"https://"

_OPEN_ID_CONFIG_PATH = u".well-known/openid-configuration"


class KeyUriSupplier(object):  # pylint: disable=too-few-public-methods
    """A supplier that provides the `jwks_uri` for an issuer."""

    def __init__(self, issuer_uri_configs):
        """Construct an instance of KeyUriSupplier.

        Args:
          issuer_uri_configs: a dictionary mapping from an issuer to its jwks_uri
            configuration.
        """
        self._issuer_uri_configs = issuer_uri_configs

    def supply(self, issuer):
        """Supplies the `jwks_uri` for the given issuer.

        Args:
          issuer: the issuer.

        Returns:
          The `jwks_uri` that is either statically configured or retrieved via
          OpenId discovery. None is returned when the issuer is unknown or the
          OpenId discovery fails.
        """
        issuer_uri_config = self._issuer_uri_configs.get(issuer)

        if not issuer_uri_config:
            # The issuer is unknown.
            return

        jwks_uri = issuer_uri_config.jwks_uri
        if jwks_uri:
            # When jwks_uri is set, return it directly.
            return jwks_uri

        # When jwksUri is empty, we try to retrieve it through the OpenID
        # discovery.
        open_id_valid = issuer_uri_config.open_id_valid
        if open_id_valid:
            discovered_jwks_uri = _discover_jwks_uri(issuer)
            self._issuer_uri_configs[issuer] = IssuerUriConfig(False,
                                                               discovered_jwks_uri)
            return discovered_jwks_uri


class JwksSupplier(object):  # pylint: disable=too-few-public-methods
    """A supplier that returns the Json Web Token Set of an issuer."""

    def __init__(self, key_uri_supplier):
        """Constructs an instance of JwksSupplier.

        Args:
          key_uri_supplier: a KeyUriSupplier instance that returns the `jwks_uri`
            based on the given issuer.
        """
        self._key_uri_supplier = key_uri_supplier
        self._jwks_cache = cache.make_region().configure(
            u"dogpile.cache.memory", expiration_time=datetime.timedelta(minutes=5))

    def supply(self, issuer):
        """Supplies the `Json Web Key Set` for the given issuer.

        Args:
          issuer: the issuer.

        Returns:
          The successfully retrieved Json Web Key Set. None is returned if the
            issuer is unknown or the retrieval process fails.

        Raises:
          UnauthenticatedException: When this method cannot supply JWKS for the
            given issuer (e.g. unknown issuer, HTTP request error).
        """
        def _retrieve_jwks():
            """Retrieve the JWKS from the given jwks_uri when cache misses."""
            jwks_uri = self._key_uri_supplier.supply(issuer)

            if not jwks_uri:
                raise UnauthenticatedException(u"Cannot find the `jwks_uri` for issuer "
                                               u"%s: either the issuer is unknown or "
                                               u"the OpenID discovery failed" % issuer)

            try:
                response = requests.get(jwks_uri)
                json_response = response.json()
            except Exception as exception:
                message = u"Cannot retrieve valid verification keys from the `jwks_uri`"
                raise UnauthenticatedException(message, exception)

            if u"keys" in json_response:
                # De-serialize the JSON as a JWKS object.
                jwks_keys = jwk.KEYS()
                jwks_keys.load_jwks(response.text)
                return jwks_keys._keys
            else:
                # The JSON is a dictionary mapping from key id to X.509 certificates.
                # Thus we extract the public key from the X.509 certificates and
                # construct a JWKS object.
                return _extract_x509_certificates(json_response)

        return self._jwks_cache.get_or_create(issuer, _retrieve_jwks)


def _extract_x509_certificates(x509_certificates):
    keys = []
    for kid, certificate in x509_certificates.iteritems():
        try:
            if certificate.startswith(jwk.PREFIX):
                # The certificate is PEM-encoded
                der = ssl.PEM_cert_to_DER_cert(certificate)
                key = jwk.der2rsa(der)
            else:
                key = jwk.import_rsa_key(certificate)
        except Exception as exception:
            raise UnauthenticatedException(u"Cannot load X.509 certificate",
                                           exception)
        rsa_key = jwk.RSAKey().load_key(key)
        rsa_key.kid = kid
        keys.append(rsa_key)
    return keys


def _discover_jwks_uri(issuer):
    open_id_url = _construct_open_id_url(issuer)
    try:
        response = requests.get(open_id_url)
        return response.json().get(u"jwks_uri")
    except Exception as error:
        raise UnauthenticatedException(u"Cannot discover the jwks uri", error)


def _construct_open_id_url(issuer):
    url = issuer
    if (not url.startswith(_HTTP_PROTOCOL_PREFIX) and
            not url.startswith(_HTTPS_PROTOCOL_PREFIX)):
        url = _HTTPS_PROTOCOL_PREFIX + url
    if not url.endswith(u"/"):
        url += u"/"
    url += _OPEN_ID_CONFIG_PATH
    return url


class IssuerUriConfig(object):
    """The jwks_uri configuration for an issuer.

    TODO (yangguan): this class should be removed after we figure out how to
    fetch the external configs.
    """

    def __init__(self, open_id_valid, jwks_uri):
        """Create an instance of IsserUriConfig.

        Args:
          open_id_valid: indicates whether the corresponding issuer is valid for
            OpenId discovery.
          jwks_uri: is the saved jwks_uri. Its value can be None if the OpenId
            discovery process has not begun or has already failed.
        """
        self._open_id_valid = open_id_valid
        self._jwks_uri = jwks_uri

    @property
    def open_id_valid(self):
        return self._open_id_valid

    @property
    def jwks_uri(self):
        return self._jwks_uri


class UnauthenticatedException(Exception):
    pass



# Copyright 2018 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test WSGI responses when check/quota errors occur."""

import mock
import pytest
import webtest
from webtest.debugapp import debug_app as DEBUG_APP

from apitools.base.py import encoding
from endpoints_management.control import (
    client, quota_request, sc_messages, sm_messages, wsgi,
)
from .test_wsgi import _SYSTEM_PARAMETER_CONFIG_TEST


@pytest.fixture(scope='module')
def project_id():
    return 'middleware-errors'


@pytest.fixture()
def control_client():
    return mock.MagicMock(spec=client.Client)


@pytest.fixture()
def service_config_loader():
    service = encoding.JsonToMessage(sm_messages.Service, _SYSTEM_PARAMETER_CONFIG_TEST)
    loader = mock.Mock()
    loader.load.return_value = service
    return loader


@pytest.fixture()
def wrapped_app(project_id, control_client, service_config_loader):
    return wsgi.add_all(DEBUG_APP, project_id, control_client, loader=service_config_loader)


@pytest.fixture()
def test_app(wrapped_app):
    return webtest.TestApp(wrapped_app, lint=False)


def test_handle_missing_api_key(control_client, test_app):
    url = '/uvw/method_needs_api_key/more_stuff'
    check_resp = sc_messages.CheckResponse(
        operationId=u'fake_operation_id')
    control_client.check.return_value = check_resp
    resp = test_app.get(url, expect_errors=True)
    assert resp.status_code == 401
    assert resp.content_type == 'application/json'
    assert wsgi.Middleware._NO_API_KEY_MSG in resp.json['message']


def test_handle_out_of_quota(control_client, test_app):
    quota_resp = sc_messages.AllocateQuotaResponse(
        allocateErrors = [
            sc_messages.QuotaError(
                code=quota_request._QuotaErrors.RESOURCE_EXHAUSTED,
                description=u'details')
        ]
    )
    check_resp = sc_messages.CheckResponse(
        operationId=u'fake_operation_id')
    control_client.check.return_value = check_resp
    control_client.allocate_quota.return_value = quota_resp
    url = '/uvw/method2/with_no_param'
    resp = test_app.get(url, expect_errors=True)
    expected_status, expected_detail = quota_request._QUOTA_ERROR_CONVERSION[
        quota_request._QuotaErrors.RESOURCE_EXHAUSTED]
    assert resp.status_code == expected_status
    assert resp.content_type == 'application/json'
    assert expected_detail in resp.json['message']
